/***************************************************************************
                           cfilemanager.cpp  -  description
                             -------------------
    begin                : Sat Aug 03 2002
    copyright            : (C) 2002-2005 by Mathias Kster
    email                : mathen@users.berlios.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef WIN32
#include <unistd.h>
#include <stdlib.h>
#endif

#include <string.h>

#include "dcos.h"
#include "core/cbytearray.h"
#include "cconfig.h"
#include "core/cstring.h"
#include "core/cstringlist.h"
#include "core/cdir.h"
#include "core/cfile.h"
#include "cdownloadmanager.h"
#include "dclib.h"
#include "core/cmanager.h"
#include "csearchindex.h"
#include "core/cbase32.h"

#include "hash/compat.h"
#include "hash/MerkleTree.h"

#include "cfilemanager.h"

class CFileType : public CObject {
public:
	/** */
	CFileType() {};
	/** */
	virtual ~CFileType() {};

	/** */
	eFileTypes m_eFileType;
};

/** filetype list */
typedef struct filetype {
	const char* sExt;
	eFileTypes eFileType;
} filetype;

struct filetype FileType[] = {
	// MP3 2
	{ "APE",  eftMP3 },
	{ "AIFF", eftMP3 },
	{ "AU",   eftMP3 },
	{ "FLA",  eftMP3 },
	{ "FLAC", eftMP3 },
	{ "MID",  eftMP3 },
	{ "M4A",  eftMP3 },
	{ "M4P",  eftMP3 },
	{ "MPC",  eftMP3 },
	{ "MP2",  eftMP3 },
	{ "MP3",  eftMP3 },
	{ "OGG",  eftMP3 },
	{ "RA",   eftMP3 },
	{ "SHN",  eftMP3 },
	{ "WAV",  eftMP3 },
	{ "WMA",  eftMP3 },
	{ "WV",   eftMP3 },

	// ARCHIVE 3
	{ "7Z",  eftARCHIVE },
	{ "ACE", eftARCHIVE },
	{ "BZ2", eftARCHIVE },
	{ "CAB", eftARCHIVE },
	{ "EX_", eftARCHIVE },
	{ "GZ",  eftARCHIVE },
	{ "LZH", eftARCHIVE },
	{ "RAR", eftARCHIVE },
	{ "RPM", eftARCHIVE },
	{ "TAR", eftARCHIVE },
	{ "TGZ", eftARCHIVE },
	{ "ZIP", eftARCHIVE },
	{ "ZOO", eftARCHIVE },
	{ "Z",   eftARCHIVE },

	// DOCUMENT 4
	{ "CFG",   eftDOCUMENT },
	{ "CONF",  eftDOCUMENT },
	{ "CPP",   eftDOCUMENT },
	{ "CSS",   eftDOCUMENT },
	{ "C",     eftDOCUMENT },
	{ "DIZ",   eftDOCUMENT },
	{ "DOC",   eftDOCUMENT },
	{ "H",     eftDOCUMENT },
	{ "HLP",   eftDOCUMENT },
	{ "HTM",   eftDOCUMENT },
	{ "HTML",  eftDOCUMENT },
	{ "INI",   eftDOCUMENT },
	{ "INF",   eftDOCUMENT },
	{ "LOG",   eftDOCUMENT },
	{ "NFO",   eftDOCUMENT },
	{ "ODG",   eftDOCUMENT },
	{ "ODP",   eftDOCUMENT },
	{ "ODS",   eftDOCUMENT },
	{ "ODT",   eftDOCUMENT },
	{ "PHP",   eftDOCUMENT },
	{ "PPT",   eftDOCUMENT },
	{ "PS",    eftDOCUMENT },
	{ "PDF",   eftDOCUMENT },
	{ "SHTML", eftDOCUMENT },
	{ "SXC",   eftDOCUMENT },
	{ "SXD",   eftDOCUMENT },
	{ "SXI",   eftDOCUMENT },
	{ "SXW",   eftDOCUMENT },
	{ "TXT",   eftDOCUMENT },
	{ "RFT",   eftDOCUMENT },
	{ "RDF",   eftDOCUMENT },
	{ "XML",   eftDOCUMENT },
	{ "XLS",   eftDOCUMENT },

	// APPL 5
	{ "BAT",  eftAPPLICATION },
	{ "CGI",  eftAPPLICATION },
	{ "COM",  eftAPPLICATION },
	{ "DLL",  eftAPPLICATION },
	{ "EXE",  eftAPPLICATION },
	{ "HQX",  eftAPPLICATION },
	{ "JS",   eftAPPLICATION },
	{ "SH",   eftAPPLICATION },
	{ "SO",   eftAPPLICATION },
	{ "SYS",  eftAPPLICATION },
	{ "VXD",  eftAPPLICATION },

	// PICTURE 6
	{ "3DS",  eftPICTURE },
	{ "A11",  eftPICTURE },
	{ "ACB",  eftPICTURE },
	{ "ADC",  eftPICTURE },
	{ "ADI",  eftPICTURE },
	{ "AFI",  eftPICTURE },
	{ "AI",   eftPICTURE },
	{ "AIS",  eftPICTURE },
	{ "ANS",  eftPICTURE },
	{ "ART",  eftPICTURE },
	{ "B8",   eftPICTURE },
	{ "BMP",  eftPICTURE },
	{ "CBM",  eftPICTURE },
	{ "GIF",  eftPICTURE },
	{ "ICO",  eftPICTURE },
	{ "JPEG", eftPICTURE },
	{ "JPG",  eftPICTURE },
	{ "PIC",  eftPICTURE },
	{ "PNG",  eftPICTURE },
	{ "PCX",  eftPICTURE },
	{ "PSP",  eftPICTURE },
	{ "TGA",  eftPICTURE },
	{ "TIF",  eftPICTURE },
	{ "XPM",  eftPICTURE },

	// VIDEO 7
	{ "AVI",   eftVIDEO },
	{ "ASF",   eftVIDEO },
	{ "ASX",   eftVIDEO },
	{ "DAT",   eftVIDEO },
	{ "FLV",   eftVIDEO },
	{ "MKV",   eftVIDEO },
	{ "MOV",   eftVIDEO },
	{ "MOVIE", eftVIDEO },
	{ "MPG",   eftVIDEO },
	{ "MP4",   eftVIDEO },
	{ "MPEG",  eftVIDEO },
	{ "OGM",   eftVIDEO },
	{ "PXP",   eftVIDEO },
	{ "QT",    eftVIDEO },
	{ "RM",    eftVIDEO },
	{ "RMVB",  eftVIDEO },
	{ "VIV",   eftVIDEO },
	{ "WMV",   eftVIDEO },
	{ "DIVX",  eftVIDEO },

	// END
	{ 0, eftALL }
};

/** */
CFileManager::CFileManager()
{
	m_pFileTypeList   = new CStringList();
	m_pFileNameList   = 0;
	m_nFileBaseIndex  = 0;
	m_pSearchIndex    = new CSearchIndex();
	m_pShareList      = new CShareList();
	m_pHashMemory     = 0;

	m_pFileManagerInfo = new CFileManagerInfo();
	m_pFileManagerInfo->m_nProgress = 100;

	InitFileTypeList();

	if ( m_pShareList->Load() == FALSE )
	{
		CreateShareList();
	}
	else
	{
		if ( m_pSearchIndex->LoadedOK() == FALSE ) // CSearchIndex::LoadIndex() is called in CSearchIndex constructor
		{
			CreateShareList();
		}
		else
		{
			m_pShareList->CreateList(m_pSearchIndex);
		}
		
		if ( CConfig::Instance()->GetAutoRecreateShareList() )
		{
			// this test was always broken before since m_pShareList->CreateList(m_pSearchIndex) had not been called
			// let's just always re-create it if the setting is true
			// it is unlikely but possible that files have changed but the total size is still the same
			//if ( CalcShareSize() != m_pShareList->GetShareSize() )
			//{
				CreateShareList();
			//}
		}
	}

	if ( CConfig::Instance()->GetRecreateShareListTime() != 0 )
	{
		m_tCreateShareListTimeout = time(0)+CConfig::Instance()->GetRecreateShareListTime()*60*60;
	}
	else
	{
		m_tCreateShareListTimeout = 0;
	}

	// init timer
	m_pFileManagerCallback = new CCallback<CFileManager>( this, &CFileManager::FileManagerCallback );
	CManager::Instance()->Add( m_pFileManagerCallback );
}

/** */
CFileManager::~CFileManager()
{
	Stop();

	SetInstance(0);

	Lock();

	if ( m_pFileManagerCallback )
	{
		CManager::Instance()->Remove(m_pFileManagerCallback);
		delete m_pFileManagerCallback;
	}

	if ( m_pFileNameList )
	{
		delete m_pFileNameList;
		m_pFileNameList = 0;
	}

	if ( m_pFileManagerInfo )
	{
		delete m_pFileManagerInfo;
		m_pFileManagerInfo = 0;
	}

	if ( m_pFileTypeList )
	{
		delete m_pFileTypeList;
		m_pFileTypeList = 0;
	}

	if ( m_pShareList )
	{
		delete m_pShareList;
		m_pShareList = 0;
	}

	if ( m_pSearchIndex )
	{
		delete m_pSearchIndex;
		m_pSearchIndex = 0;
	}

	if ( m_pHashMemory )
	{
		delete m_pHashMemory;
		m_pHashMemory = 0;
	}

	UnLock();
}

/** */
int CFileManager::GetShareBuffer( eShareBufferType type, CByteArray * sharebuffer, bool decompress )
{
	return m_pShareList->GetShareBuffer( type, sharebuffer, decompress );
}

/** */
unsigned long CFileManager::GetShareBufferSize( eShareBufferType type )
{
	return m_pShareList->GetShareBufferSize(type);
}

/** */
ulonglong CFileManager::GetShareSize()
{
	if ( m_pShareList )
		return m_pShareList->GetShareSize();
	else
		return 0;
}

/** */
void CFileManager::InitFileTypeList()
{
	int i;
	CFileType * filetype;

	for(i=0;FileType[i].sExt != 0;i++)
	{
		filetype = new CFileType();
		filetype->m_eFileType = FileType[i].eFileType;

		m_pFileTypeList->Add( FileType[i].sExt, filetype );
	}
}

/** return the filetype */
eFileTypes CFileManager::GetFileType( CString file )
{
	CString ext;
	eFileTypes res=eftUNKNOWN;
	CFileType * filetype;

	// set the filetype
	ext = CDir::Extension(file);

	if ( ext != "" )
	{
		if ( m_pFileTypeList->Get( ext.ToUpper(), (CObject*&)filetype ) == 0 )
		{
			res = filetype->m_eFileType;
		}
#if 0
		if ( res == eftUNKNOWN )
			printf("unknown extension: %s\n",ext.Data());
#endif
	}

	return res;
}

/** */
CStringList * CFileManager::Search( CString s )
{
	if ( !m_pFileManagerInfo || !m_pSearchIndex )
		return 0;

	//if ( m_pFileManagerInfo->m_eFileManagerStatus != efmsNONE )
	//	return 0;

	return m_pSearchIndex->Search(s);
}

/** */
CStringList * CFileManager::SearchHash( CString s )
{
	CString ts;
	CByteArray dst;
	CBase32 base32;
	
	if ( !m_pFileManagerInfo || !m_pSearchIndex )
		return 0;

	//if ( m_pFileManagerInfo->m_eFileManagerStatus != efmsNONE )
	//	return 0;

	if ( s.Left(4) != "TTH:" )
		return 0;

	ts = s.Mid(4,s.Length()-4);
	
	if ( base32.Decode(&dst,&ts) != TigerTree::HASH_SIZE )
		return 0;

	return m_pSearchIndex->SearchHash(dst.Data());
}

/** */
CString CFileManager::GetHash( ulonglong hbi )
{
	if ( !m_pFileManagerInfo || !m_pSearchIndex )
		return "";

	//if ( m_pFileManagerInfo->m_eFileManagerStatus != efmsNONE )
	//	return "";
	
	return m_pSearchIndex->GetHash(hbi);
}

/** */
bool CFileManager::GetFileBaseObject( CString id, struct filebaseobject * fbo, CString & filename )
{
	if ( !m_pFileManagerInfo || !m_pSearchIndex )
		return FALSE;

	//if ( m_pFileManagerInfo->m_eFileManagerStatus != efmsNONE )
	//	return FALSE;

	return m_pSearchIndex->GetFileBaseObject(id,fbo,filename);
}

/** */
bool CFileManager::CreateShareList()
{
	bool err = FALSE;

	if ( m_pFileManagerInfo->m_eFileManagerStatus != efmsNONE )
		return err;
	
	Lock();

	if ( Start() == -1 )
	{
		UnLock();
		return err;
	}
	
	m_SharedFolders.Clear();

	m_pShareFolder = 0;
	m_nShareSize   = 0;
	m_sShareIndexBuffer = "";

	m_pSearchIndex->PrepareUpdate();

	if ( CConfig::Instance()->GetSharedFolders(&m_SharedFolders) > 0 )
	{
		if ( m_pFileNameList )
			delete m_pFileNameList;
		m_pFileNameList  = new CStringList(25);

		m_pFileManagerInfo->m_nProgress = 0;
		m_pFileManagerInfo->m_eFileManagerStatus = efmsCREATESHARELIST;

		if ( CDownloadManager::Instance() )
			CDownloadManager::Instance()->SendFileManagerInfo(m_pFileManagerInfo);

		err = TRUE;
	}
	else
	{
		m_pSearchIndex->FinishUpdate();
		
		m_pShareList->SetIndexBuffer( "" );
		m_pShareList->CreateList(m_pSearchIndex);

		m_pSearchIndex->SaveIndex();
		
		UnLock();
		Stop();
		
		return err;
	}

	UnLock();

	return err;
}

/** */
void CFileManager::CreateSearchIndex()
{
	if ( m_pFileManagerInfo->m_eFileManagerStatus == efmsIDLE )
	{
		m_nFileBaseIndex = 0;

		m_pFileManagerInfo->m_nProgress = 0;
		m_pFileManagerInfo->m_eFileManagerStatus = efmsCREATESEARCHINDEX;

		Start();
	}
}

/** */
bool CFileManager::CreateHashList()
{
	bool res = FALSE;

	if ( m_pFileManagerInfo->m_eFileManagerStatus == efmsIDLE )
	{
		m_nFileBaseIndex = 0;

		m_pFileManagerInfo->m_nProgress = 0;
		m_pFileManagerInfo->m_eFileManagerStatus = efmsCREATEHASHLIST;

		Start();
		
		res = TRUE;
	}

	return res;
}

/** */
void CFileManager::ThreadCreateSearchIndex()
{
	int i,k;
	CString s,s1,id;
	double percent;

	for(k=0;k<100;k++)
	{
		if ( (s = m_pSearchIndex->GetFileNameDuringUpdate(m_nFileBaseIndex)) != "" )
		{
			s = s.ToUpper();

			m_pSearchIndex->AddIndex(s,m_nFileBaseIndex);

			// convert special chars
			s = s.Replace(' ',"/");
			s = s.Replace('.',"/");
			s = s.Replace('-',"/");
			s = s.Replace('_',"/");
			s = s.Replace('(',"/");
			s = s.Replace(')',"/");
			s = s.Replace('!',"/");

			// DIRSEPARATOR is added on the end so that the
			// last word gets indexed. Usually this is the
			// file extension, but for a directory it will
			// be part of the folder name - ejs1920
			s = CDir().ConvertSeparators(s);
			s += DIRSEPARATOR;

			while( (i=s.Find(DIRSEPARATOR)) != -1 )
			{
				s1 = s.Left(i);
				s  = s.Mid(i+1,s.Length()-i-1);

				i++;

				// max length is 3
				if ( s1.Length() < 3 )
					continue;

				m_pSearchIndex->AddIndex(s1,m_nFileBaseIndex);
			}

			m_nFileBaseIndex++;
		}
		else
		{
			m_pSearchIndex->InitIndex();

			//delete m_pSearchArray;
			//m_pSearchArray = 0;

			printf("ready create index\n");

			// not yet, only do it once after hash list refresh finished
			//m_pSearchIndex->SaveIndex();

			m_pFileManagerInfo->m_nProgress = 100;
			m_pFileManagerInfo->m_eFileManagerStatus = efmsIDLE;

			if ( CDownloadManager::Instance() )
				CDownloadManager::Instance()->SendFileManagerInfo(m_pFileManagerInfo);

			CreateHashList();

			return;
		}
	}

	if ( m_pSearchIndex->IndexCountDuringUpdate() > 0 )
	{
		percent = (m_nFileBaseIndex*100.0)/m_pSearchIndex->IndexCountDuringUpdate();

		if ( m_pFileManagerInfo->m_nProgress != percent )
		{
			m_pFileManagerInfo->m_nProgress = percent;

			if ( CDownloadManager::Instance() )
				CDownloadManager::Instance()->SendFileManagerInfo(m_pFileManagerInfo);
		}
	}
}

/** */
void CFileManager::ThreadCreateShareList()
{
	CString s,s1;
	CDir dir;

	if ( (m_pShareFolder=m_SharedFolders.Next(m_pShareFolder)) != 0 )
	{
		// directory
		s = m_pShareFolder->m_sPath;

		// is valid ?
		if ( dir.cd(s) == TRUE )
		{
			s  = dir.Path();
			s1 = dir.DirName();

			if ( s1 != "" )
			{
				s = s.Left( s.Length()-s1.Length() );
			}

			ThreadCreateShareList( 0, "", "", 100/m_SharedFolders.Count() );
		}
		else
		{
			printf("Can't change dir: '%s'\n",s.Data());
		}
	}
	else
	{
		m_pFileManagerInfo->m_nProgress = 100;
		printf("ready create sharelist\n");

		m_pShareList->SetIndexBuffer( m_sShareIndexBuffer );
		
		// m_pShareList->CreateList will be called after hashing finishes
		// don't call it here to avoid serving up filelists without TTHs in
		//m_pShareList->CreateList(m_pSearchIndex);

		if ( m_pFileNameList )
		{
			delete m_pFileNameList;
			m_pFileNameList = 0;
		}

		m_pFileManagerInfo->m_eFileManagerStatus = efmsIDLE;

		if ( CDownloadManager::Instance() )
			CDownloadManager::Instance()->SendFileManagerInfo(m_pFileManagerInfo);

		CreateSearchIndex();

		if ( CDownloadManager::Instance() )
			CDownloadManager::Instance()->SendFileManagerInfo(m_pFileManagerInfo);
	}
}

/** */
void CFileManager::ThreadCreateShareList( int depth, CString curr, CString rel, double percent )
{
	CDir dir;
	CList<CFileInfo> list;
	CString d,s,*ts=0;
	CString path;
	CFileInfo *fileinfo=0;
	int x,z;
	unsigned long index;

	if ( Stopped() )
		return;

	d = m_pShareFolder->m_sPath;
	d += DIRSEPARATOR;
	d += curr;
	d += DIRSEPARATOR;
	d += rel;
	path = dir.SimplePath(d);
	
	if ( (CConfig::Instance()->GetDontShareDotFiles() == TRUE) && (rel.Left(1) == ".") )
	{
		DPRINTF( "Not sharing folder \"%s\" due to dot files setting.\n", path.Data() );
		return;
	}

	dir.SetPath(path);

	d = "";
	z = depth;

//	printf("%s\n",curr.Data());

	// create depth string
	for(x=depth;x>0;x--) d+="\x9";
	z++;
	// use alias
	if ( depth == 0 )
	{
		m_sShareIndexBuffer += m_pShareFolder->m_sAlias;
		m_sShareIndexBuffer += "\xd\xa";
	}
	else
	{
		m_sShareIndexBuffer += d;
		m_sShareIndexBuffer += rel;
		m_sShareIndexBuffer += "\xd\xa";
	}
	d += "\x9";

	// add folder so it can be a search result
	fileinfo = new CFileInfo();
	
	if ( depth == 0 )
	{
		fileinfo->name = m_pShareFolder->m_sAlias;
	}
	else
	{
		fileinfo->name = rel;
	}
	fileinfo->size = 0;
	fileinfo->m_bSymlink = FALSE;
	fileinfo->st_a_time = 0;
	fileinfo->st_m_time = 0;
	fileinfo->st_c_time = 0;
	
	if ( depth == 0 )
	{
		m_pSearchIndex->AddIndex( fileinfo, DIRSEPARATOR, eftFOLDER );
	}
	else
	{
		m_pSearchIndex->AddIndex( fileinfo, dir.SimplePath(m_pShareFolder->m_sAlias+DIRSEPARATOR+curr), eftFOLDER );
	}
	
	delete fileinfo;
	fileinfo = 0;

	// get all files
	if ( dir.ReadEntrys( CDir::Files, &list ) )
	{
		// add files
		fileinfo = 0;
		while( (fileinfo=list.Next(fileinfo)) != 0 )
		{
			if ( (CConfig::Instance()->GetDontShareDotFiles() == TRUE) && (fileinfo->name.Left(1) == ".") )
			{
				DPRINTF( "Not sharing file \"%s\" due to dot files setting.\n", fileinfo->name.Data() );
				continue;
			}
			
			if ( Stopped() )
				return;

			if ( fileinfo->name != "" )
			{
				if ( fileinfo->m_bSymlink )
					s = dir.ReadLink(path+DIRSEPARATOR+fileinfo->name);
				else
					s = dir.SimplePath(path+DIRSEPARATOR+fileinfo->name);

				if ( s != "" )
				{
					if ( m_pFileNameList->Get(s,(CObject*&)ts) != 0 )
					{
						if ( dclibVerbose() > 1 )
							printf("CreateShareList: Add file: '%s/%s'\n",path.Data(),fileinfo->name.Data());fflush(stdout);

						m_pFileNameList->Add(s,new CString());
						m_nShareSize   += fileinfo->size;
						
						// add file to the file manager ...
						index = m_pSearchIndex->AddIndex( fileinfo, dir.SimplePath(m_pShareFolder->m_sAlias+DIRSEPARATOR+curr+DIRSEPARATOR+rel), GetFileType(fileinfo->name) );
						
						m_sShareIndexBuffer += d;
						m_sShareIndexBuffer += fileinfo->name;
						m_sShareIndexBuffer += '|';
						m_sShareIndexBuffer += CString().setNum(index);
						m_sShareIndexBuffer += "\xd\xa";
					}
				}
			}
		}
	}

	// get all dirs
	if ( dir.ReadEntrys( CDir::Dirs, &list ) )
	{
		if ( list.Count() > 0 )
			percent /= list.Count();
		else
			m_pFileManagerInfo->m_nProgress += percent;

		// add dir
		fileinfo = 0;
		while( (fileinfo=list.Next(fileinfo)) != 0 )
		{
			if ( Stopped() )
				return;

			if ( (fileinfo->name != ".") && (fileinfo->name != "..") )
			{
				if ( fileinfo->m_bSymlink )
					s = dir.ReadLink(path+DIRSEPARATOR+fileinfo->name);
				else
					s = dir.SimplePath(path+DIRSEPARATOR+fileinfo->name);

				if ( s != "" )
				{
					if ( m_pFileNameList->Get(s,(CObject*&)ts) != 0 )
					{
						if ( dclibVerbose() > 0 )
							printf("CreateShareList: Add path: '%s/%s'\n",path.Data(),fileinfo->name.Data());fflush(stdout);

						m_pFileNameList->Add(s,new CString());
						ThreadCreateShareList(z,curr+DIRSEPARATOR+rel,fileinfo->name,percent);
					}
					else
						m_pFileManagerInfo->m_nProgress += percent;

				}
			}
			else
				m_pFileManagerInfo->m_nProgress += percent;

			if ( CDownloadManager::Instance() )
				CDownloadManager::Instance()->SendFileManagerInfo(m_pFileManagerInfo);
		}
	}
}

/** */
ulonglong CFileManager::CalcShareSize()
{
	CString s,s1;
	CDir dir;
	ulonglong iSSize;

	iSSize = 0;
	m_pShareFolder=0;

	if ( CConfig::Instance()->GetSharedFolders(&m_SharedFolders) == 0 )
	{
		printf("No share folderfound !");
		return 0;
	}

	while ( (m_pShareFolder=m_SharedFolders.Next(m_pShareFolder)) != 0 )
	{
		// directory
		s = m_pShareFolder->m_sPath;

		// is valid ?
		if ( dir.cd(s) == TRUE )
		{
			s  = dir.Path();
			s1 = dir.DirName();

			if ( s1 != "" )
			{
				s = s.Left( s.Length()-s1.Length() );
			}

			iSSize += CalcShareSize( 0, s, s1, "" );
		}
		else
		{
			printf("Can't change to dir: '%s'\n",s.Data());
		}
	}

	return iSSize;
}


/** */
ulonglong CFileManager::CalcShareSize( int depth, CString base, CString curr, CString relpath )
{
	ulonglong lShareSize;
	CDir dir;
	CList<CFileInfo> list;
	CString d,s;
	CString path;
	CString rpath;
	CFileInfo *fileinfo;
	int x,z;

	lShareSize = 0;

	if ( relpath == "" )
	{
		rpath = curr;
	}
	else
	{
		rpath = relpath;
		rpath += DIRSEPARATOR;
		rpath += curr;
	}

	if (base=="")
	{
		path = curr;
	}
	else
	{
		path = base;
		path += DIRSEPARATOR;
		path += curr;
	}

	dir.SetPath(path);

	d = "";
	z = depth;

	// create depth string
	if ( curr != "" )
	{
		for(x=depth;x>0;x--) d+="\x9";
		z++;
		d += "\x9";
	}

	// get all files
	if ( dir.ReadEntrys( CDir::Files, &list ) )
	{
		fileinfo = 0;
		while( (fileinfo=list.Next(fileinfo)) != 0 )
		{
			if ( fileinfo->name != "" )
			{
				if ( fileinfo->m_bSymlink )
					s = dir.ReadLink(path+DIRSEPARATOR+fileinfo->name);
				else
					s = dir.SimplePath(path+DIRSEPARATOR+fileinfo->name);

				if ( s != "" )
				{
					lShareSize += fileinfo->size;
				}
			}
		}
	}

	// get all dirs
	if ( dir.ReadEntrys( CDir::Dirs, &list ) )
	{
		fileinfo = 0;
		while( (fileinfo=list.Next(fileinfo)) != 0 )
		{
			if ( (fileinfo->name != ".") && (fileinfo->name != "..") )
			{
				if ( fileinfo->m_bSymlink )
					s = dir.ReadLink(path+DIRSEPARATOR+fileinfo->name);
				else
					s = dir.SimplePath(path+DIRSEPARATOR+fileinfo->name);

				if ( s != "" )
				{
					lShareSize += CalcShareSize(z,path,fileinfo->name,rpath);
				}
			}
		}
	}
	return lShareSize;
}

/** */
void CFileManager::ThreadCreateHashList()
{
	CString s,pa;
	struct filebaseobject fbo;
	ulonglong hbi;
	int len;
	CFile f;
	CDir cdir;
	double percent;

	if ( !m_pHashMemory )
	{
		m_pHashMemory = new CByteArray();
		m_pHashMemory->SetSize(1024*1024);
	}
	
	if ( m_pSearchIndex->GetFileBaseObjectDuringUpdate(CString().setNum(m_nFileBaseIndex),&fbo,s) == TRUE )
	{
		pa = CConfig::Instance()->AliasToPath(s);
		
		if ( pa == "" )
		{
			m_nFileBaseIndex++;
			return;
		}
		
		// check if we have the hash
		if ( m_pSearchIndex->FindHashBaseIndexDuringUpdate(&fbo,&hbi) )
		{
			if ( dclibVerbose() > 0 )
				printf("hash found, no changes in file %s\n", pa.Data());
			fbo.m_nHashIndex = hbi;
			m_pSearchIndex->UpdateIndex(m_nFileBaseIndex,&fbo);
		}
		else
		{
			if ( CConfig::Instance()->GetDisableHashList() == FALSE )
			{
			
				if ( cdir.IsDir( pa, FALSE ) == FALSE )
				{
				
				if ( dclibVerbose() > 0 )
					printf("calculate new hash for %s\n", pa.Data());
				
				if ( f.Open( pa, IO_RAW | IO_READONLY ) == TRUE )
				{
					TigerTree tth( max( TigerTree::calcBlockSize(cdir.getFileSize(pa,FALSE),10) , 64*1024 ) );

					while ( (len = f.Read( (char*)m_pHashMemory->Data(), 1024*1024)) > 0 )
					{
						tth.update(m_pHashMemory->Data(),len);
				
						if ( Stopped() )
							break;
					}
					
					f.Close();
			
					if ( !Stopped() )
					{
						tth.finalize();
						vector<uint8_t> a = tth.getLeafData();
						m_pSearchIndex->AddHashIndex(m_nFileBaseIndex,
							tth.getRoot().data,
							&a[0],a.size());
					}
					else
					{
						printf("create hash stop\n");
						m_pFileManagerInfo->m_eFileManagerStatus = efmsIDLE;
					}
				}
				
				}
			}
		}
		
		m_nFileBaseIndex++;
	
		if ( m_pSearchIndex->IndexCountDuringUpdate() > 0 )
		{
			percent = (m_nFileBaseIndex*100.0)/m_pSearchIndex->IndexCountDuringUpdate();

			if ( (percent-m_pFileManagerInfo->m_nProgress) > 0.5 )
			{
				m_pFileManagerInfo->m_nProgress = percent;

				if ( CDownloadManager::Instance() )
					CDownloadManager::Instance()->SendFileManagerInfo(m_pFileManagerInfo);
			}
		}
	}
	else
	{
		m_pFileManagerInfo->m_eFileManagerStatus = efmsIDLE;
	}
	
	if ( m_pFileManagerInfo->m_eFileManagerStatus == efmsIDLE )
	{
		printf("create hash end\n");
		m_pFileManagerInfo->m_nProgress = 100;

		if ( CDownloadManager::Instance() )
			CDownloadManager::Instance()->SendFileManagerInfo(m_pFileManagerInfo);

		if ( m_pHashMemory )
		{
			delete m_pHashMemory;
			m_pHashMemory = 0;
		}

		m_pSearchIndex->FinishUpdate();
		m_pSearchIndex->SaveIndex();
		
		// SaveIndex() closes and saves hash leaves
		//m_pSearchIndex->CloseHashLeaves();
		
		m_pShareList->CreateList(m_pSearchIndex);
		
		m_pFileManagerInfo->m_eFileManagerStatus = efmsNONE;
		if ( CDownloadManager::Instance() )
			CDownloadManager::Instance()->SendFileManagerInfo(m_pFileManagerInfo);

		Stop();
	}
}

/** */
void CFileManager::Thread( CObject * )
{
	int wait = 50;

	Lock();

	switch(m_pFileManagerInfo->m_eFileManagerStatus)
	{
		case efmsCREATESHARELIST:
			ThreadCreateShareList();
			break;
		case efmsCREATESEARCHINDEX:
			ThreadCreateSearchIndex();
			wait = 0;
			break;
		case efmsCREATEHASHLIST:
			ThreadCreateHashList();
			wait = 0;
			break;
		case efmsREBUILDLISTS:
			ThreadRebuildLists();
			wait = 0;
			break;
		default:
			break;
	}

	UnLock();

	NanoSleep(wait);
}

/** */
int CFileManager::FileManagerCallback( CObject *, CObject * )
{
	if ( CConfig::Instance()->GetRecreateShareListTime() != 0 )
	{
		if ( m_tCreateShareListTimeout != 0 )
		{
			if ( time(0) > m_tCreateShareListTimeout )
			{
				CreateShareList();

				m_tCreateShareListTimeout = time(0)+CConfig::Instance()->GetRecreateShareListTime()*60*60;
			}
		}
		else
		{
			m_tCreateShareListTimeout = time(0)+CConfig::Instance()->GetRecreateShareListTime()*60*60;
		}
	}
	else
	{
		m_tCreateShareListTimeout = 0;
	}

	return 0;
}

/** */
CString CFileManager::GetFileName( CString fboid )
{
	struct filebaseobject fbo;
	CString filename;

	if ( GetFileBaseObject(fboid,&fbo,filename) )
		return filename;
	return "";
}

/** */
bool CFileManager::IsCreateShareListRunning()
{
	if (m_pFileManagerInfo->m_eFileManagerStatus == efmsNONE)
	{
		return FALSE;
	}
	else
	{
		return TRUE;
	}
}

/** */
CByteArray * CFileManager::GetHashLeaves( CString tth )
{
	return m_pSearchIndex->GetHashLeaves( tth );
}

/** */
bool CFileManager::RebuildLists()
{
	bool err = FALSE;
	
	//DPRINTF("CFileManager::RebuildLists() called\n");
	
	if ( m_pFileManagerInfo->m_eFileManagerStatus != efmsNONE )
	{
		return err;
	}
	
	Lock();
	
	m_pFileManagerInfo->m_eFileManagerStatus = efmsREBUILDLISTS;
	
	Start();
	
	err = TRUE;
	
	UnLock();
	
	return err;
}

/** */
void CFileManager::ThreadRebuildLists()
{
	//DPRINTF("CFileManager::ThreadRebuildLists() called\n");
	
	// CSearchIndex has direct access to the byte arrays so do it there
	m_pSearchIndex->RebuildLists();
	
	m_pFileManagerInfo->m_nProgress = 0;
	m_pFileManagerInfo->m_eFileManagerStatus = efmsCREATESHARELIST;
	
	// recreate search index (also redoes hash lists)
	// needed bits copied out of CFileManager::CreateShareList
	// the mutex is already locked
	// the thread is already running
	// we already know m_eFileManagerStatus
	
	m_pShareFolder = 0;
	m_nShareSize   = 0;
	m_sShareIndexBuffer = "";
	
	m_pSearchIndex->PrepareUpdate();
	
	if ( CConfig::Instance()->GetSharedFolders(&m_SharedFolders) > 0 )
	{
		if ( m_pFileNameList )
			delete m_pFileNameList;
		m_pFileNameList = new CStringList(25);
		
		if ( CDownloadManager::Instance() )
			CDownloadManager::Instance()->SendFileManagerInfo(m_pFileManagerInfo);	
	}
	else
	{
		m_pSearchIndex->FinishUpdate();
		
		m_pShareList->SetIndexBuffer( "" );
		m_pShareList->CreateList(m_pSearchIndex);

		m_pSearchIndex->SaveIndex();
		
		UnLock();
		Stop();
	}
	//DPRINTF("CFileManager::ThreadRebuildLists() finished\n");
}
