/***************************************************************************
                          cstring.cpp  -  description
                             -------------------
    begin                : Fri Sep 21 2001
    copyright            : (C) 2001-2003 by Mathias Kster
    email                : mathen@users.berlios.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "../dcos.h"
#include "cstring.h"

CString::CString()
{
	Init();
}

CString::CString( const char * sz )
{
	Init();

	set(sz);
}

CString::CString( const char ch )
{
	Init();

	set(&ch,1);
}

CString::CString( const CString & stringSrc ) : CObject()
{
	Init();

	set(stringSrc.m_szBuffer,stringSrc.m_nStringLength);
}

CString::~CString()
{
	if (m_szBuffer)
	{
		free(m_szBuffer);
		m_szBuffer = 0;
	}
}

/** */
void CString::Init()
{
	m_szBuffer      = 0;
	m_nStringLength = 0;
	m_nBufferSize   = 0;
}

/** */
void CString::Empty()
{
	if (m_szBuffer)
	{
		free(m_szBuffer);
	}

	Init();
}

/** */
unsigned char CString::GetHash( long value ) const
{
	if ( (!IsEmpty()) && (value < m_nStringLength) )
		return ((unsigned char)(m_szBuffer[value]&0xFF));
	else
		return 0;
}

/** */
long CString::Find( const char ch, long nStart ) const
{
	long i;

	if ( IsEmpty() )
	{
		return -1;
	}

	if ( nStart > m_nStringLength )
		return -1;

	for(i=nStart;i<m_nStringLength;i++)
	{
		if ( m_szBuffer[i] == ch )
			return i;
	}

	return -1;
}

/** */
long CString::Find( const char * sz, long nStart, bool cs ) const
{
	/* long i; */
	long len;
	long ret;

	if ( !cs )
	{
		return FindCase(sz,nStart);
	}

	if ( IsEmpty() )
	{
		if ( sz != 0 )
			return -1;
		else
			return 0;
	}

	if ( !sz )
		return -1;

	len = strlen(sz);

	if ( (nStart+len) > m_nStringLength )
		return -1;

	/* for(i=nStart,ret=-1;i<=(m_nStringLength-len);i++)
	{
		if ( strncmp(sz,m_szBuffer+i,len) == 0 )
		{
			ret = i; //-index;
			break;
		}
	} */
	
	char * location = strstr( m_szBuffer + nStart, sz );
	
	if ( location == NULL )
	{
		ret = -1;
	}
	else
	{
		ret = location - m_szBuffer;
	}

	return ret;
}

/** */
long CString::Find( const CString & string, long nStart, bool cs ) const
{
	/* long i; */
	long len;
	long ret;
	char * sz;

	if ( !cs )
	{
		return FindCase(string,nStart);
	}

	sz  = string.Data();
	len = string.Length();

	if ( IsEmpty() )
	{
		if ( sz != 0 )
			return -1;
		else
			return 0;
	}

	if ( !sz )
		return -1;

	if ( (nStart+len) > m_nStringLength )
		return -1;

	/* for(i=nStart,ret=-1;i<=(m_nStringLength-len);i++)
	{
		if ( strncmp(sz,m_szBuffer+i,len) == 0 )
		{
			ret = i; //-index;
			break;
		}
	} */

	char * location = strstr( m_szBuffer + nStart, sz );
	
	if ( location == NULL )
	{
		ret = -1;
	}
	else
	{
		ret = location - m_szBuffer;
	}

	return ret;
}

/** */
long CString::FindCase( const char * sz, long nStart ) const
{
	long len;
	long ret;

	if ( IsEmpty() )
	{
		if ( sz != 0 )
			return -1;
		else
			return 0;
	}

	if ( !sz )
		return -1;

	len = strlen(sz);

	if ( (nStart+len) > m_nStringLength )
		return -1;

#ifdef HAVE_STRCASESTR

	char * location = strcasestr( m_szBuffer + nStart, sz );
	
	if ( location == NULL )
	{
		ret = -1;
	}
	else
	{
		ret = location - m_szBuffer;
	}

#else // HAVE_STRCASESTR

	long i;
	for(i=nStart,ret=-1;i<=(m_nStringLength-len);i++)
	{
#ifdef WIN32
		if ( _strnicmp(sz,m_szBuffer+i,len) == 0 )
#else
		if ( strncasecmp(sz,m_szBuffer+i,len) == 0 )
#endif
		{
			ret = i;
			break;
		}
	}

#endif // HAVE_STRCASESTR

	return ret;
}

/** */
long CString::FindCase( const CString & string, long nStart ) const
{
	long len;
	long ret;
	char * sz;

	sz  = string.Data();
	len = string.Length();

	if ( IsEmpty() )
	{
		if ( sz != 0 )
			return -1;
		else
			return 0;
	}

	if ( !sz )
		return -1;

	if ( (nStart+len) > m_nStringLength )
		return -1;
#ifdef HAVE_STRCASESTR

	char * location = strcasestr( m_szBuffer + nStart, sz );
	
	if ( location == NULL )
	{
		ret = -1;
	}
	else
	{
		ret = location - m_szBuffer;
	}

#else // HAVE_STRCASESTR

	long i;
	for(i=nStart,ret=-1;i<=(m_nStringLength-len);i++)
	{
#ifdef WIN32
		if ( _strnicmp(sz,m_szBuffer+i,len) == 0 )
#else
		if ( strncasecmp(sz,m_szBuffer+i,len) == 0 )
#endif
		{
			ret = i;
			break;
		}
	}

#endif // HAVE_STRCASESTR

	return ret;
}

/** */
long CString::FindRev( char ch, long nStart ) const
{
	long i;

	if ( IsEmpty() )
		return -1;

	if ( nStart == -1 )
		nStart = m_nStringLength;

	if ( nStart > (m_nStringLength+1) )
		return -1;

	for(i=nStart;i>=0;i--)
	{
		if (m_szBuffer[i]==ch)
			return i;
	}

	return -1;
}

/** */
long CString::FindRev( const CString & string ) const
{
	long i;
	long len;
	long ret=-1;
	char * sz;

	sz  = string.Data();
	len = string.Length();

	if ( IsEmpty() )
	{
		if ( sz != 0 )
			return -1;
		else
			return 0;
	}

	if ( !sz )
		return -1;

	for(i=(m_nStringLength-len);i>=0;i--)
	{
		if ( string == CString(m_szBuffer).Mid(i,len) )
		{
			ret = i; //-index;
			break;
		}
	}

	return ret;
}

/** */
CString CString::Mid( long nFirst, long nCount ) const
{
	CString ret;

	// added_by_frg for nFirst < 0 case:
	// which is needed in all occasions where we use str.Right(n) where n > str.Length()
	if (nFirst < 0)
	{
		nCount += nFirst;
		nFirst = 0;
	}
	// addition_end

	if ( IsEmpty() )
		return ret;
	if ( nCount == -1 )
		nCount = m_nStringLength-nFirst;
	if ( nCount <= 0 )
		return ret;
	if( nFirst > m_nStringLength )
		return ret;

	if( (nFirst+nCount) > m_nStringLength )
		return ret;

	ret.set(m_szBuffer+nFirst,nCount);

	return ret;
}

/** */
CString CString::Section( const char sep, int start, int end ) const
{
	int i,si,ei;

	if ( IsEmpty() )
		return CString();

	for(i=0,si=0;i<start;i++,si++)
	{
		si = Find(sep,si);
		if ( si == -1 )
			break;
	}

	if ( si == -1 )
		return CString();

	for(ei=si;i<=end;i++,ei++)
	{
		ei = Find(sep,ei);
		if ( ei == -1 )
			break;
	}

	if ( ei == -1 )
		ei = m_nStringLength + 1;

	/* if ( (si==-1) || (ei==-1) )
		return CString(); */

	return Mid(si,ei-si-1);
}

/** */
bool CString::IsEmpty() const
{
	return ( (m_nStringLength==0) || (m_szBuffer==0) );
}

/** */
CString CString::setNum( const int n )
{
	char c[50];

#ifdef WIN32
	_snprintf(c,50,"%d",n);
#else
	snprintf(c,50,"%d",n);
#endif

	return CString(c);
}

/** */
CString CString::setNum( const unsigned int n )
{
	char c[50];

#ifdef WIN32
	_snprintf(c,50,"%u",n);
#else
	snprintf(c,50,"%u",n);
#endif

	return CString(c);
}

/** */
CString CString::setNum( const long n )
{
	char c[50];

#ifdef WIN32
	_snprintf(c,50,"%ld",n);
#else
	snprintf(c,50,"%ld",n);
#endif

	return CString(c);
}

/** */
CString CString::setNum( const unsigned long n )
{
	char c[50];

#ifdef WIN32
	_snprintf(c,50,"%lu",n);
#else
	snprintf(c,50,"%lu",n);
#endif

	return CString(c);
}

/** */
CString CString::setNum( const long long n )
{
	char c[50];

#ifdef WIN32
	_snprintf(c,50,"%lld",n);
#else
	snprintf(c,50,"%lld",n);
#endif

	return CString(c);
}

/** */
CString CString::setNum( const ulonglong n )
{
	char c[50];

#ifdef WIN32
	_ui64toa( n, c, 10 );
#else
	snprintf(c,50,"%llu",n);
#endif

	return CString(c);
}

/** */
CString CString::setNum( const double n, const int p )
{
	char c[50];

#ifdef WIN32
	_snprintf(c,50,"%.*f",p,n);
#else
	snprintf(c,50,"%.*f",p,n);
#endif
	return CString(c);
}

/** */
ulonglong CString::asULL( int base ) const
{
	ulonglong res;
	char *tail=0;

	if ( IsEmpty() )
	{
		return 0;
	}

#ifdef WIN32
	res=strtoull(Data(),&tail,base);
#else
	#ifdef HAVE_STRTOULL
	res=strtoull(Data(),&tail,base);
	#elif HAVE_STRTOUQ
	res=strtouq(Data(),&tail,base);
	#elif defined(__arch64__) 
 	res=strtoul(Data(),&tail,base);
	#else
	#error "You don't have strtoull or strtouq function !"
	#endif
#endif

	if((res==0)&&(errno!=0)) {}

	return res;
}

/** */
unsigned int CString::asUINT( int base ) const
{
	return (unsigned int)asLONG(base);
}

/** */
int CString::asINT( int base ) const
{
	return (int)asLONG(base);
}

/** */
long CString::asLONG( int base ) const
{
	unsigned int res;
	char *tail=0;

	if ( IsEmpty() )
	{
		return 0;
	}

	res=strtol(Data(),&tail,base);

	if((res==0)&&(errno!=0)) {}

	return res;
}

long long CString::asLONGLONG( int base ) const
{
	long long res;
	char *tail=0;
	
	if ( IsEmpty() )
	{
		return 0;
	}
	
	res=strtoll(Data(),&tail,base);
	
	if((res==0)&&(errno!=0)) {}
	
	return res;
}

/** */
double CString::asDOUBLE() const
{
	char *endptr = 0;

	if ( IsEmpty() )
	{
		return 0;
	}

	return strtod(Data(),&endptr);
}

/** */
CString CString::ToUpper() const
{
	CString ret="";
	long i=0;

	if ( IsEmpty() )
		return ret;

	for(;i<m_nStringLength;i++)
	{
		ret += (const char)(toupper(m_szBuffer[i])&0xFF);
	}

	return ret;
}

/** */
CString CString::ToLower() const
{
	CString ret="";
	long i=0;

	if ( IsEmpty() )
		return ret;

	for(;i<m_nStringLength;i++)
	{
		ret += (const char)(tolower(m_szBuffer[i])&0xFF);
	}

	return ret;
}

/** */
CString CString::Replace( CString src, CString string ) const
{
	long i,t;
	CString ret;

	t = i = 0;
	while ( (i=Find(src.Data(),i)) != -1 )
	{
		ret += Mid(t,i-t);
		ret += string;
		i+=src.Length();
		t = i;
	}

	ret += Mid(t,m_nStringLength-t);

	return ret;
}

/** */
void CString::set( const char * sz, long nLength )
{
	long l;
	char * temp;

	if ( !sz )
	{
		Empty();
		return;
	}

	if( nLength == -1 )
		l = strlen(sz);
	else
		l = nLength;

	if ( l <= 0 )
	{
		Empty();
		return;
	}

	if ( (m_szBuffer) && (sz >= m_szBuffer) && ( sz <= m_szBuffer+m_nBufferSize ) )
	{
		temp = (char*) malloc( l );
		if ( temp == 0 )
		{
			perror("CString::set malloc");
			return;
		}
		memcpy(temp,sz,l);
	}
	else
	{
		temp = (char*) sz;
	}

	Empty();

	m_szBuffer = (char*) malloc(l+1); //new char[l+1];

	if ( m_szBuffer == 0 )
	{
		printf("CString::set malloc [%ld]: %s\n",l+1,strerror(errno));
		
		if ( temp != sz )
		{
			free(temp);
		}
		
		return;
	}

	memcpy(m_szBuffer,temp,l);
	m_szBuffer[l]   = 0;
	m_nStringLength = l;
	m_nBufferSize   = l+1;
	
	if ( temp != sz )
	{
		free(temp);
	}
}

/** */
void CString::add( const char * sz, long nLength )
{
	long l,len,i;
	char *p;
	char * temp;

	if (!sz)
		return;

	if( nLength == -1 )
		l = strlen(sz);
	else
		l = nLength;

	if (l<=0)
		return;

	if(!m_szBuffer)
	{
		set(sz,l);
	}
	else
	{
		// if pointers overlap, copy input before realloc
		// which also means memcpy is safe (do not need memmove)
		if ( (sz >= m_szBuffer) && (sz <= m_szBuffer+m_nBufferSize) )
		{
			temp = (char*) malloc( l );
			if ( temp == 0 )
			{
				perror("CString::add malloc");
				return;
			}
			memcpy(temp,sz,l);
		}
		else
		{
			temp = (char*) sz;
		}

		len = m_nStringLength;

		if((l+len+1)>m_nBufferSize)
		{
			i = 1000+1+l;
			p = (char*) realloc(m_szBuffer,m_nBufferSize+i);

			if ( p == 0 )
			{
				perror("CString::add realloc");
				
				if ( temp != sz )
				{
					free(temp);
				}
				
				return;
			}

			m_szBuffer = p;
			m_nBufferSize += i;
		}

		memset(m_szBuffer+len+l,0,1);
		memcpy(m_szBuffer+len,temp,l);
		m_nStringLength+=l;
		
		if ( temp != sz )
		{
			free(temp);
		}
	}
}

#ifdef WIN32
// original from dietlibc
ulonglong CString::strtoull(const char *nptr, char **endptr, int base) const
{
  longlong v=0;

  while(isspace(*nptr)) ++nptr;

  if (*nptr == '+') ++nptr;
  if (!base) {
    if (*nptr=='0') {
      base=8;
      if ((*(nptr+1)=='x')||(*(nptr+1)=='X')) {
	nptr+=2;
	base=16;
      }
    }
    else
      base=10;
  }
  while(*nptr) {
    register unsigned char c=*nptr;
    c=(c>='a'?c-'a'+10:c>='A'?c-'A'+10:c-'0');
    if (c>=base) break;
    v=v*base+c;
    ++nptr;
  }
  if (endptr) *endptr=(char *)nptr;
  return v;
}
#endif

/** */
bool operator == ( const CString & string1, const CString & string2 )
{
	if ( string1.IsEmpty() && string2.IsEmpty() )
		return TRUE;

	if ( string1.Length() != string2.Length() )
		return FALSE;

	if ( memcmp( string1.Data(), string2.Data(), string1.Length() ) != 0 )
		return FALSE;

	return TRUE;
}

/** */
CString CString::RightJustify( long nNewLength, char fill , bool /*truncate*/ )
{
	CString s="";

	while( (s.m_nStringLength+m_nStringLength) < nNewLength )
		s+=fill;

	return(s+*this);
}

/** */
bool operator == ( const char * sz, const CString & string )
{
	if ( sz == 0 )
	{
		if ( string.IsEmpty() )
		{
			return TRUE;
		}
		else
		{
			return FALSE;
		}
	}
	else
	{
		if ( strlen(sz) == (size_t) string.Length() )
		{
			return ( memcmp( sz, string.Data(), string.Length() ) == 0 );
		}
		else
		{
			return FALSE;
		}
	}
}

/** */
bool operator == ( const CString & string, const char * sz )
{
	return ( sz == string );
}

/** */
bool operator != ( const CString & string1, const CString & string2 )
{
	return (!(string1 == string2));
}

/** */
bool operator != ( const char * sz, const CString & string )
{
	return (!(sz == string));
}

/** */
bool operator != ( const CString & string, const char * sz )
{
	return (!(sz == string));
}

/** */
int CString::Compare( const CString & other ) const
{
	bool this_is_empty = IsEmpty();
	bool other_is_empty = other.IsEmpty();
	
	if ( this_is_empty && other_is_empty )
	{
		return 0;
	}
	else if ( this_is_empty && !other_is_empty )
	{
		return 1;
	}
	else if ( !this_is_empty && other_is_empty )
	{
		return -1;
	}
	else
	{
		return strcoll( Data(), other.Data() );
	}
}

/** */
int CString::Compare( const char * other ) const
{
	bool this_is_empty = IsEmpty();
	bool other_is_empty = (other == 0);
	
	if ( this_is_empty && other_is_empty )
	{
		return 0;
	}
	else if ( this_is_empty && !other_is_empty )
	{
		return 1;
	}
	else if ( !this_is_empty && other_is_empty )
	{
		return -1;
	}
	else
	{
		return strcoll( Data(), other );
	}
}
