/***************************************************************************
                           cxml.cpp  -  description
                             -------------------
    begin                : Wed May 15 2002
    copyright            : (C) 2002-2005 by Mathias Kster
    email                : mathen@users.berlios.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#define __STDC_ISO_10646__  200104L
 
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>

// for strlen, which is used instead of creating a temp CString
#include <string.h>

#include "ciconv.h"

#include "cxml.h"

/** */
CXml::CXml()
{
	int i;
	
	pDoc = 0;

	CString local = getenv("LANG");
	i = local.Find('.');
	if ( i == -1 )
	{
#ifdef __APPLE__
		local = "UTF-8";
#else
		local = "ISO-8859-1"; // this is the old fallback but probably should also be UTF-8
#endif
	}
	else
	{
		local = local.Mid(i + 1);
	}
	
	pToUTF8   = new CIconv( local, "UTF-8" );
	pFromUTF8 = new CIconv( "UTF-8", local );
}

/** */
CXml::~CXml()
{
	FreeDoc();

	delete pToUTF8;
	delete pFromUTF8;
}

/** */
void CXml::FreeDoc()
{
	if (pDoc)
	{
		xmlFreeDoc(pDoc);
		pDoc = 0;
	}
}

/** */
bool CXml::NewDoc()
{
	FreeDoc();

	if ( (pDoc = xmlNewDoc((const xmlChar*)"1.0")) == 0 )
	{
		return FALSE;
	}

	return TRUE;
}

/** */
bool CXml::ParseFile( CString name )
{
	FreeDoc();

	if ( (pDoc = xmlRecoverFile(name.Data())) == 0 )
	{
		return FALSE;
	}

	return TRUE;
}

/** */
bool CXml::ParseMemory( const char * s, int size )
{
	FreeDoc();

	if ( (pDoc = xmlRecoverMemory(s,size)) == 0 )
	{
		return FALSE;
	}

	return TRUE;
}

/** */
CString CXml::content( xmlNodePtr node )
{
	xmlChar *c;
	CString s = "";

	if ( (c = xmlNodeGetContent(node)) != 0 )
	{
		s = FromUtf8((char*)c);
		xmlFree(c);
	}

	return s;
}

/** */
CString CXml::prop( xmlNodePtr node, const CString & prop )
{
	xmlChar *c;
	CString s = "";

	if ( (c = xmlGetProp(node,(const xmlChar*)prop.Data())) != 0 )
	{
		s = (char*)c;
		xmlFree(c);
	}

	return s;
}

/** */
CString CXml::prop( xmlNodePtr node, const char * prop )
{
	xmlChar *c;
	CString s = "";

	if ( (c = xmlGetProp(node,(const xmlChar*)prop)) != 0 )
	{
		s = (char*)c;
		xmlFree(c);
	}

	return s;
}

/** */
CString CXml::ToUTF8( const char * s )
{
	if ( (s == 0) || (s[0] == 0) )
	{
		return CString();
	}
	else
	{
		CString r = pToUTF8->encode(s);
		
		if ( xmlCheckUTF8( (const unsigned char*) r.Data() ) == FALSE )
		{
			printf("CXml::ToUTF8: iconv returned invalid UTF-8, doing ISO-8859-1 to UTF-8\n");
			printf("input='%s' (char *) iconv='%s'\n",s,r.Data());
			
			int inlen = strlen(s);
			int outlen = inlen * 4;
			
			unsigned char * b = (unsigned char*) calloc(1,outlen);
			
			if ( b != 0 )
			{
				int res = isolat1ToUTF8( b, &outlen, (unsigned char*)s, &inlen );
				
				if ( res >= 0 )
					r = (char*)b;
				else
					printf("CXml::ToUTF8 isolat1ToUTF8 fail: '%s'\n",s);
				
				free(b);
			}
		}
		
		return EscapeSpecials(r);
	}
}

/** */
CString CXml::FromUtf8( const char * s )
{
	if ( (s == 0) || (s[0] == 0) )
	{
		return CString();
	}
	else
	{
		return UnEscapeSpecials( pFromUTF8->encode(s) );
	}
}

/** */
CString CXml::ToUTF8( const CString & s )
{
	if ( s.IsEmpty() )
	{
		return CString();
	}
	else
	{
		CString r = pToUTF8->encode(s);
		
		if ( xmlCheckUTF8( (const unsigned char*) r.Data() ) == FALSE )
		{
			printf("CXml::ToUTF8: iconv returned invalid UTF-8, doing ISO-8859-1 to UTF-8\n");
			printf("input='%s' (CString) iconv='%s'\n",s.Data(),r.Data());
			
			int inlen = s.Length();
			int outlen = inlen * 4;
			
			unsigned char * b = (unsigned char*) calloc(1,outlen);
			
			if ( b != 0 )
			{
				int res = isolat1ToUTF8( b, &outlen, (unsigned char*)s.Data(), &inlen );
				
				if ( res >= 0 )
					r = (char*)b;
				else
					printf("CXml::ToUTF8 isolat1ToUTF8 fail: '%s'\n",s.Data());
				
				free(b);
			}
		}
		
		return EscapeSpecials(r);
	}
}

/** */
CString CXml::FromUtf8( const CString & s )
{
	if ( s.IsEmpty() )
	{
		return CString();
	}
	else
	{
		return UnEscapeSpecials( pFromUTF8->encode(s) );
	}

}

CString CXml::EscapeSpecials( const CString & s )
{
	int i;
	CString dst1;
	
	if ( s.IsEmpty() )
	{
		return dst1;
	}

	// sorry, CString::Replace is too slow
	/* dst = s.Replace("&","&amp;");
	dst = dst.Replace("\'","&apos;");
	dst = dst.Replace("\"","&quot;");
	dst = dst.Replace("<","&lt;");
	dst = dst.Replace(">","&gt;"); */

	/* 0x00 - 0x08, 0x0B, 0x0C, 0x0E - 0x1F have been declared to
	 * be illegal in XML documents
	*/
	for(i=0;i<s.Length();i++)
	{
		int x = s.Data()[i]&0xff;
		
		if ( ((x <= 0x08) && (x >= 0x00)) ||
		     (x == 0x0B) ||
		     (x <= 0x0C) ||
		     ((x <= 0x1F) && (x >= 0x0E)) )
		{
			dst1 += "&#0";
			dst1 += CString().setNum(x);
			dst1 += ';';
		}
		else if ( x == '&' )
		{
			dst1 += "&amp;";
		}
		else if ( x == '\'' )
		{
			dst1 += "&apos;";
		}
		else if ( x == '"' )
		{
			dst1 += "&quot;";
		}
		else if ( x == '<' )
		{
			dst1 += "&lt;";
		}
		else if ( x == '>' )
		{
			dst1 += "&gt;";
		}
		else
		{
			dst1 += s.Data()[i];
		}
	}

	return dst1;

}

CString CXml::UnEscapeSpecials( const CString & s )
{
	int i,i1;
	CString res;
	CString temp;
	CString four,six;

	for(i=0;i<s.Length();i++)
	{
		if ( s.Data()[i] == '&' )
		{
			if ( s.Mid(i,2) == "&#" )
			{
				if ( (i1 = s.Find(';',i)) != -1 )
				{
					if ( (i1-i) <= 5 )
					{
						temp = s.Mid(i,i1-i);				
						//printf("%s\n",t.Data());
						
						temp = temp.Mid(2);
						
						if ( temp.Left(1) == "x" )
						{
							temp = temp.Mid(1);
							
							res += temp.asINT(16);
						}
						else
						{
							res += temp.asINT();
						}
						
						//printf("%s\n",temp.Data());
						i += i1-i;
						continue;
					}
				}
			}
			else
			{
				four = s.Mid(i,4);
				if ( four == "&lt;" )
				{
					res += '<';
					i += 3;
					continue;
				}
				else if ( four == "&gt;" )
				{
					res += '>';
					i += 3;
					continue;
				}
				else
				{
					if ( s.Mid(i,5) == "&amp;" )
					{
						res += '&';
						i += 4;
						continue;
					}
					else
					{
						six = s.Mid(i,6);
						if ( six == "&apos;" )
						{
							res += '\'';
							i += 5;
							continue;
						}
						else if ( six == "&quot;" )
						{
							res += '"';
							i += 5;
							continue;
						}
					}
				}
			}
		}
		
		res += s.Data()[i];
	}

	// too slow, probably
	/* res = res.Replace("&apos;", "\'");
	res = res.Replace("&quot;", "\"");
	res = res.Replace("&lt;",   "<");
	res = res.Replace("&gt;",   ">");
	res = res.Replace("&amp;",  "&"); */

	return res;
}

/** */
xmlNodePtr CXml::xmlNewBoolChild( xmlNodePtr parent, xmlNsPtr ns, const xmlChar *name, bool b )
{
	if (b)
		return xmlNewChild( parent, ns, name, (const xmlChar*)(const char*)XML_TRUE );
	else
		return xmlNewChild( parent, ns, name, (const xmlChar*)(const char*)XML_FALSE );
}

/** */
xmlNodePtr CXml::xmlNewStringChild( xmlNodePtr parent, xmlNsPtr ns, const xmlChar *name, const CString & s, bool createempty )
{
	if ( (createempty==TRUE) || (s!="") )
		return xmlNewTextChild( parent, ns, name, (const xmlChar*)ToUTF8(s).Data() );
	else
		return 0;
}

/** */
xmlNodePtr CXml::xmlNewStringChild( xmlNodePtr parent, xmlNsPtr ns, const xmlChar *name, const char * s, bool createempty )
{
	if ( (createempty==TRUE) || (s != 0) )
		return xmlNewTextChild( parent, ns, name, (const xmlChar*)ToUTF8(s).Data() );
	else
		return 0;
}

/** */
void CXml::xmlNewStringProp( xmlNodePtr node, const CString & prop, const CString & value )
{
	xmlNewProp( node, (const xmlChar*)prop.Data(), (const xmlChar*)ToUTF8(value).Data() );
}

/** */
void CXml::xmlNewStringProp( xmlNodePtr node, const char * prop, const char * value )
{
	xmlNewProp( node, (const xmlChar*)prop, (const xmlChar*)ToUTF8(value).Data() );
}

/** */
bool CXml::getBoolChild( xmlNodePtr node )
{
	if ( content(node) == XML_TRUE )
		return TRUE;
	else
		return FALSE;
}

/** */
void CXml::InitParser()
{
	if ( xmlParserVersion != CString("20510") )
	{
		xmlInitParser();
	}
}

/** */
void CXml::CleanupParser()
{
#if LIBXML_VERSION > 20510
	if ( xmlParserVersion != CString("20510") )
	{
		xmlCleanupParser();
	}
#endif
}
