/***************************************************************************
                           cxml.h  -  description
                             -------------------
    begin                : Sun Jun 9 2002
    copyright            : (C) 2002-2004 by Mathias Kster
    email                : mathen@users.berlios.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CXML_H
#define CXML_H

/**
  *@author Mathias Kster
  */

#include <dclib/dcos.h>
#include <dclib/core/cobject.h>
#include <dclib/core/cstring.h>

#include <libxml/xmlversion.h>
#if LIBXML_VERSION > 20406
#include <libxml/globals.h>
#endif
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

#define XML_FALSE			"false"
#define XML_TRUE			"true"

class CIconv;

class DLL_EXPORT CXml : public CObject {

public:
	/** */
	CXml();
	/** */
	virtual ~CXml();

	/** */
	xmlDocPtr doc();
	/** */
	CString name( xmlNodePtr node );
	/** */
	CString content( xmlNodePtr node );
	/** */
	CString prop( xmlNodePtr node, const CString & prop );
	/** */
	CString prop( xmlNodePtr node, const char * prop );
	/** */
	xmlNodePtr xmlNewBoolChild( xmlNodePtr parent, xmlNsPtr ns, const xmlChar *name, bool b );
	/** */
	xmlNodePtr xmlNewStringChild( xmlNodePtr parent, xmlNsPtr ns, const xmlChar *name, const CString & s, bool createempty = TRUE );
	/** */
	xmlNodePtr xmlNewStringChild( xmlNodePtr parent, xmlNsPtr ns, const xmlChar *name, const char * s, bool createempty = TRUE );
	/** */
	void xmlNewStringProp( xmlNodePtr node, const CString & prop, const CString & value );
	/** */
	void xmlNewStringProp( xmlNodePtr node, const char * prop, const char * value );

	/** */
	bool getBoolChild( xmlNodePtr node );

	/** */
	bool NewDoc();
	/** */
	bool ParseFile( CString name );
	/** */
	bool ParseMemory( const char * s, int size );

	/** */
	CString ToUTF8( const CString & s );
	/** */
	CString FromUtf8( const CString & s );
	/** */
	CString ToUTF8( const char * s );
	/** */
	CString FromUtf8( const char * s );
	/** */
	CString EscapeSpecials( const CString & s );
	/** */
	CString UnEscapeSpecials( const CString & s );
	
	/** Must be called once on program startup */
	static void InitParser();
	/** Must be called once on program shutdown */
	static void CleanupParser();

private:
	/** */
	void FreeDoc();

	/** */
	xmlDocPtr pDoc;
	/** Encodes from local to UTF-8 */
	CIconv * pToUTF8;
	/** Encodes from UTF-8 to local */
	CIconv * pFromUTF8;
};

inline xmlDocPtr CXml::doc()
{ return pDoc; }
inline CString CXml::name( xmlNodePtr node )
{ return CString(((char*)node->name)); }

#endif
