/***************************************************************************
                           csearchindex.cpp  -  description
                             -------------------
    begin                : Mon May 14 2003
    copyright            : (C) 2003-2004 by Mathias Kster
    email                : mathen@users.berlios.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef WIN32
#include <unistd.h>
#include <stdlib.h>
#endif

#include <string.h>

#include "dcos.h"
#include "core/cbytearray.h"
#include "cconfig.h"
#include "core/cstring.h"
#include "core/cstringlist.h"
#include "core/cdir.h"
#include "core/cfile.h"
#include "cdownloadmanager.h"
#include "dclib.h"
#include "core/cmanager.h"
#include "cfilemanager.h"
#include "core/cbase32.h"

#include "hash/compat.h"
#include "hash/MerkleTree.h"

#include "cfilehasher.h"

#include "csearchindex.h"

/** */
CSearchIndex::CSearchIndex()
{
	m_pBaseArray       = new CByteArray();
	m_pFileBaseArray   = new CByteArray();
	m_pPathBaseArray   = new CByteArray();
	
	m_pSearchIndex     = new CByteArray();
	m_pSearchFileIndex = new CByteArray();
	m_pSearchBase      = new CByteArray();

	m_pHashBaseArray = new CByteArray();
	m_pHashFileBaseArray = new CByteArray();
	m_pHashPathBaseArray = new CByteArray();
	m_pHashIndex = new CByteArray();
	
	m_pNewHashLeaves = 0;
	hashleavessize = 0; // this will get set when m_pNewHashLeaves is created
	
	memset(m_pSearchArray,0,sizeof(m_pSearchArray));

	m_nSearchFileIndex = 0;

	/* during list refresh, these are created and updated
	 * and replace the old byte arrays when refresh stage is finished
	 */
	m_pUpdatingFileBaseArray     = 0;
	m_pUpdatingPathBaseArray     = 0;
	m_pUpdatingBaseArray         = 0;
	
	m_pUpdatingSearchIndex       = 0;
	m_pUpdatingSearchFileIndex   = 0;
	m_pUpdatingSearchBase        = 0;
	m_nUpdatingSearchFileIndex   = 0;
	
	m_pUpdatingHashBaseArray     = 0;
	m_pUpdatingHashFileBaseArray = 0;
	m_pUpdatingHashPathBaseArray = 0;
	m_pUpdatingHashIndex         = 0;

	m_bLoadedOK = LoadIndex();
}

/** */
CSearchIndex::~CSearchIndex()
{
	delete m_pBaseArray;
	delete m_pFileBaseArray;
	delete m_pPathBaseArray;
	
	delete m_pSearchIndex;
	delete m_pSearchFileIndex;
	delete m_pSearchBase;
	
	delete m_pHashBaseArray;
	delete m_pHashFileBaseArray;
	delete m_pHashPathBaseArray;
	delete m_pHashIndex;
	
	if (m_pNewHashLeaves != 0)
	{
		DPRINTF("~CSearchIndex: warning new hash leaves not saved!\n");
		SaveNewHashLeaves();
		
		if ( m_pNewHashLeaves != 0 ) // something went wrong in SaveNewHashLeaves()
		{
			delete m_pNewHashLeaves;
		}
	}
	
	if ( m_pUpdatingFileBaseArray ) delete m_pUpdatingFileBaseArray;
	if ( m_pUpdatingPathBaseArray ) delete m_pUpdatingPathBaseArray;
	if ( m_pUpdatingBaseArray ) delete m_pUpdatingBaseArray;
	
	if ( m_pUpdatingSearchIndex ) delete m_pUpdatingSearchIndex;
	if ( m_pUpdatingSearchFileIndex ) delete m_pUpdatingSearchFileIndex;
	if ( m_pUpdatingSearchBase ) delete m_pUpdatingSearchBase;
	
	if ( m_pUpdatingHashBaseArray ) delete m_pUpdatingHashBaseArray;
	if ( m_pUpdatingHashFileBaseArray ) delete m_pUpdatingHashFileBaseArray;
	if ( m_pUpdatingHashPathBaseArray ) delete m_pUpdatingHashPathBaseArray;
	if ( m_pUpdatingHashIndex ) delete m_pUpdatingHashIndex;
}

/** */
void CSearchIndex::PrepareUpdate()
{
	// files list
	if ( m_pUpdatingFileBaseArray ) delete m_pUpdatingFileBaseArray;
	if ( m_pUpdatingPathBaseArray ) delete m_pUpdatingPathBaseArray;
	if ( m_pUpdatingBaseArray ) delete m_pUpdatingBaseArray;
	
	m_pUpdatingFileBaseArray = new CByteArray(0);
	m_pUpdatingPathBaseArray = new CByteArray(0);
	m_pUpdatingBaseArray     = new CByteArray(0);
	
	// search indexes
	if ( m_pUpdatingSearchIndex ) delete m_pUpdatingSearchIndex;
	if ( m_pUpdatingSearchFileIndex ) delete m_pUpdatingSearchFileIndex;
	if ( m_pUpdatingSearchBase ) delete m_pUpdatingSearchBase;
	
	m_pUpdatingSearchIndex     = new CByteArray(0);
	m_pUpdatingSearchFileIndex = new CByteArray(0);
	m_pUpdatingSearchBase      = new CByteArray(0);
	
	memset(m_pSearchArray,0,sizeof(m_pSearchArray));
	m_nUpdatingSearchFileIndex = 0;
	
	// hash lists
	if ( m_pUpdatingHashBaseArray ) delete m_pUpdatingHashBaseArray;
	if ( m_pUpdatingHashFileBaseArray ) delete m_pUpdatingHashFileBaseArray;
	if ( m_pUpdatingHashPathBaseArray ) delete m_pUpdatingHashPathBaseArray;
	if ( m_pUpdatingHashIndex ) delete m_pUpdatingHashIndex;
	
	m_pUpdatingHashBaseArray = new CByteArray(0);
	m_pUpdatingHashFileBaseArray = new CByteArray(0);
	m_pUpdatingHashPathBaseArray = new CByteArray(0);
	m_pUpdatingHashIndex = new CByteArray(0);
	
	m_pUpdatingHashBaseArray->Append( m_pHashBaseArray->Data(), m_pHashBaseArray->Size() );
	m_pUpdatingHashFileBaseArray->Append( m_pHashFileBaseArray->Data(), m_pHashFileBaseArray->Size() );
	m_pUpdatingHashPathBaseArray->Append( m_pHashPathBaseArray->Data(), m_pHashPathBaseArray->Size() );
	m_pUpdatingHashIndex->Append( m_pHashIndex->Data(), m_pHashIndex->Size() );
}

/** */
void CSearchIndex::FinishUpdate()
{
	CByteArray * old1  = m_pFileBaseArray;
	CByteArray * old2  = m_pPathBaseArray;
	CByteArray * old3  = m_pBaseArray;
	CByteArray * old4  = m_pSearchIndex;
	CByteArray * old5  = m_pSearchFileIndex;
	CByteArray * old6  = m_pSearchBase;
	CByteArray * old7  = m_pHashBaseArray;
	CByteArray * old8  = m_pHashFileBaseArray;
	CByteArray * old9  = m_pHashPathBaseArray;
	CByteArray * old10 = m_pHashIndex;
	
	// swap to new lists / append new data - this section needs to be as fast as possible
	// essentially all operations need to be done simultaneously
	// it used to have some data appending, but now
	// it's just 10 pointer swaps, so it should be OK
	m_pFileBaseArray     = m_pUpdatingFileBaseArray;
	m_pPathBaseArray     = m_pUpdatingPathBaseArray;
	m_pBaseArray         = m_pUpdatingBaseArray;
	m_pSearchIndex       = m_pUpdatingSearchIndex;
	m_pSearchFileIndex   = m_pUpdatingSearchFileIndex;
	m_pSearchBase        = m_pUpdatingSearchBase;
	m_nSearchFileIndex   = m_nUpdatingSearchFileIndex;
	m_pHashFileBaseArray = m_pUpdatingHashFileBaseArray;
	m_pHashPathBaseArray = m_pUpdatingHashPathBaseArray;
	m_pHashIndex         = m_pUpdatingHashIndex;
	m_pHashBaseArray     = m_pUpdatingHashBaseArray;
	// end of speed critical section
	
	m_pUpdatingFileBaseArray     = 0;
	m_pUpdatingPathBaseArray     = 0;
	m_pUpdatingBaseArray         = 0;
	m_pUpdatingSearchIndex       = 0;
	m_pUpdatingSearchFileIndex   = 0;
	m_pUpdatingSearchBase        = 0;
	m_nUpdatingSearchFileIndex   = 0;
	m_pUpdatingHashBaseArray     = 0;
	m_pUpdatingHashFileBaseArray = 0;
	m_pUpdatingHashPathBaseArray = 0;
	m_pUpdatingHashIndex         = 0;

	delete old1;
	delete old2;
	delete old3;
	delete old4;
	delete old5;
	delete old6;
	delete old7;
	delete old8;
	delete old9;
	delete old10;
}

/** */
void CSearchIndex::ResetHashIndex()
{
	m_pHashBaseArray->SetSize(0);
	m_pHashFileBaseArray->SetSize(0);
	m_pHashPathBaseArray->SetSize(0);
	m_pHashIndex->SetSize(0);
	
	if ( m_pNewHashLeaves == 0 )
	{
		m_pNewHashLeaves = new CByteArray();
	}
	m_pNewHashLeaves->SetSize(0);
	m_pNewHashLeaves->SaveToFile(CConfig::Instance()->GetConfigPath()+"hashleaves.bin");
	delete m_pNewHashLeaves;
	m_pNewHashLeaves = 0;
}

/** */
CStringList * CSearchIndex::Search( CString s )
{
	CString *ps;
	CStringList *sl=0;
	struct fileindexobject *pfio;
	struct searchindexobject * psio;

	if ( (psio = SearchIndex(s) ) != 0 )
	{
		sl = new CStringList();

		pfio = (struct fileindexobject*) (m_pSearchIndex->Data()+psio->m_nFileIndex);

		while(pfio)
		{
			ps = new CString( CString().setNum(pfio->m_nFileBaseIndex));
			sl->Add( *ps, ps );

			if ( pfio->m_nNext == 0 )
			{
				pfio = 0;
			}
			else
			{
				pfio = (struct fileindexobject*) (m_pSearchIndex->Data()+pfio->m_nNext);
			}
		}
	}

	return sl;
}

/** */
CStringList * CSearchIndex::SearchHash( unsigned char * hash )
{
	ulonglong hbi,hi,bi;
	CStringList *sl=0;
	CString *ps;

	hi = 0;
	
	while ( FindHash( hash, &hi ) == TRUE )
	{
		if ( HashBaseIndexFromHashIndex( hi, &hbi ) )
		{
			if ( BaseIndexFromHashBaseIndex( hbi, &bi ) )
			{
				if ( !sl )
					sl = new CStringList();
				ps = new CString( CString().setNum(bi));
				sl->Add( *ps, ps );
			}
		}
		hi+=TigerTree::HASH_SIZE;
	}

	return sl;
}

/** */
bool CSearchIndex::GetFileBaseObject( ulonglong index, struct filebaseobject * fbo )
{
	bool res = FALSE;

	if ( (index*sizeof(struct filebaseobject)) < m_pBaseArray->Size() )
	{
		memcpy( fbo, m_pBaseArray->Data()+(index*sizeof(struct filebaseobject)), sizeof(struct filebaseobject) );

		res = TRUE;
	}

	return res;
}

/** */
bool CSearchIndex::GetFileBaseObjectDuringUpdate( ulonglong index, struct filebaseobject * fbo )
{
	bool res = FALSE;

	if ( (index*sizeof(struct filebaseobject)) < m_pUpdatingBaseArray->Size() )
	{
		memcpy( fbo, m_pUpdatingBaseArray->Data()+(index*sizeof(struct filebaseobject)), sizeof(struct filebaseobject) );

		res = TRUE;
	}

	return res;
}

/** */
bool CSearchIndex::GetFileBaseObject( CString id, struct filebaseobject * fbo, CString & filename )
{
	bool res = FALSE;
	ulonglong index;

	index = id.asULL();

	if ( GetFileBaseObject(index,fbo) == TRUE )
	{
		if ( m_pPathBaseArray->Size() > fbo->m_nPathIndex )
		{
			filename = (char*)m_pPathBaseArray->Data()+fbo->m_nPathIndex;

			if ( filename != "" )
			{
				filename += DIRSEPARATOR;
			}
		}

		if ( m_pFileBaseArray->Size() > fbo->m_nFileIndex )
		{
			filename += (char*)m_pFileBaseArray->Data()+fbo->m_nFileIndex;
		}

		res = TRUE;
	}

	return res;
}

/** */
bool CSearchIndex::GetFileBaseObjectDuringUpdate( CString id, struct filebaseobject * fbo, CString & filename )
{
	bool res = FALSE;
	ulonglong index;

	index = id.asULL();

	if ( GetFileBaseObjectDuringUpdate(index,fbo) == TRUE )
	{
		if ( m_pUpdatingPathBaseArray->Size() > fbo->m_nPathIndex )
		{
			filename = (char*)m_pUpdatingPathBaseArray->Data()+fbo->m_nPathIndex;

			if ( filename != "" )
			{
				filename += DIRSEPARATOR;
			}
		}

		if ( m_pUpdatingFileBaseArray->Size() > fbo->m_nFileIndex )
		{
			filename += (char*)m_pUpdatingFileBaseArray->Data()+fbo->m_nFileIndex;
		}

		res = TRUE;
	}

	return res;
}

/** */
bool CSearchIndex::LoadIndex()
{
	bool err = FALSE;
	CDir d;
	ulonglong filesize;

	// first try to load hash index
	if ( m_pHashBaseArray->LoadFromFile(CConfig::Instance()->GetConfigPath()+"hashbase.bin") == FALSE )
	{
		err = TRUE;
	}
	if ( err == FALSE )
	{
		if ( m_pHashFileBaseArray->LoadFromFile(CConfig::Instance()->GetConfigPath()+"hashfilebase.bin") == FALSE )
		{
			err = TRUE;
		}
	}
	if ( err == FALSE )
	{
		if ( m_pHashPathBaseArray->LoadFromFile(CConfig::Instance()->GetConfigPath()+"hashpathbase.bin") == FALSE )
		{
			err = TRUE;
		}
	}
	if ( err == FALSE )
	{
		if ( m_pHashIndex->LoadFromFile(CConfig::Instance()->GetConfigPath()+"hashindex.bin") == FALSE )
		{
			err = TRUE;
		}
	}
	/* if ( err == FALSE )
	{
		if ( OpenHashLeaves() == FALSE )
		{
			err = TRUE;
		}
		else
		{
			CloseHashLeaves();
		} */
		
		/* Just check hashleaves.bin is present and not zero bytes */
		/* Don't even bother doing that anymore, it may be zero bytes if nothing shared or hashing disabled */
		/* if ( (CConfig::Instance()->GetDisableHashList() == FALSE) &&
		     (d.getFileSize(CConfig::Instance()->GetConfigPath()+"hashleaves.bin",FALSE) <= 0)
		   )
		{
			err = TRUE;
		}
	} */
	
	if ( err == TRUE )
	{
		ResetHashIndex();
		
		err = FALSE;
	}
	
	// sanity check
	filesize = d.getFileSize(CConfig::Instance()->GetConfigPath()+"database.bin",FALSE);

	if ( (filesize%sizeof(struct filebaseobject)) != 0 )
	{
		err = TRUE;
	}

	// load the filelist
	if ( err == FALSE )
	{
		if ( m_pBaseArray->LoadFromFile(CConfig::Instance()->GetConfigPath()+"database.bin") == FALSE )
		{
			err = TRUE;
		}
	}

	if ( err == FALSE )
	{
		if ( m_pFileBaseArray->LoadFromFile(CConfig::Instance()->GetConfigPath()+"filebase.bin") == FALSE )
		{
			err = TRUE;
		}
	}

	if ( err == FALSE )
	{
		if ( m_pPathBaseArray->LoadFromFile(CConfig::Instance()->GetConfigPath()+"pathbase.bin") == FALSE )
		{
			err = TRUE;
		}
	}

	if ( err )
	{
		m_pBaseArray->SetSize(0);
		m_pFileBaseArray->SetSize(0);
		m_pPathBaseArray->SetSize(0);
	}

	if ( err == FALSE )
	{
		if ( m_pSearchBase->LoadFromFile(CConfig::Instance()->GetConfigPath()+"searchbase.bin") == FALSE )
		{
			err = TRUE;
		}
	}

	if ( err == FALSE )
	{
		if ( m_pSearchIndex->LoadFromFile(CConfig::Instance()->GetConfigPath()+"searchindex.bin") == FALSE )
		{
			err = TRUE;
		}
	}

	if ( err == FALSE )
	{
		if ( m_pSearchFileIndex->LoadFromFile(CConfig::Instance()->GetConfigPath()+"searchfileindex.bin") == FALSE )
		{
			err = TRUE;
		}
	}

	if ( err )
	{
		m_pSearchBase->SetSize(0);
		m_pSearchIndex->SetSize(0);
		m_pSearchFileIndex->SetSize(0);
		m_nSearchFileIndex = 0;

		// TODO: recreate index
	}
	else
	{
		m_nSearchFileIndex = (struct long_s*)m_pSearchFileIndex->Data();
	}

	return !err;
}

/** */
void CSearchIndex::SaveIndex()
{
	m_pSearchBase->SaveToFile(CConfig::Instance()->GetConfigPath()+"searchbase.bin");
	m_pSearchIndex->SaveToFile(CConfig::Instance()->GetConfigPath()+"searchindex.bin");
	m_pSearchFileIndex->SaveToFile(CConfig::Instance()->GetConfigPath()+"searchfileindex.bin");
	m_pBaseArray->SaveToFile(CConfig::Instance()->GetConfigPath()+"database.bin");
	m_pFileBaseArray->SaveToFile(CConfig::Instance()->GetConfigPath()+"filebase.bin");
	m_pPathBaseArray->SaveToFile(CConfig::Instance()->GetConfigPath()+"pathbase.bin");
	
	m_pHashBaseArray->SaveToFile(CConfig::Instance()->GetConfigPath()+"hashbase.bin");
	m_pHashFileBaseArray->SaveToFile(CConfig::Instance()->GetConfigPath()+"hashfilebase.bin");
	m_pHashPathBaseArray->SaveToFile(CConfig::Instance()->GetConfigPath()+"hashpathbase.bin");
	m_pHashIndex->SaveToFile(CConfig::Instance()->GetConfigPath()+"hashindex.bin");
	
	if (m_pNewHashLeaves != 0)
	{
		SaveNewHashLeaves();
	}
}

/** */
unsigned long CSearchIndex::IndexCount()
{
	unsigned long i = 0;

	if ( m_pBaseArray )
	{
		i = m_pBaseArray->Size()/sizeof(struct filebaseobject);
	}

	return i;
}

/** */
unsigned long CSearchIndex::IndexCountDuringUpdate()
{
	unsigned long i = 0;

	if ( m_pUpdatingBaseArray )
	{
		i = m_pUpdatingBaseArray->Size()/sizeof(struct filebaseobject);
	}

	return i;
}

/** */
unsigned long CSearchIndex::AddIndex( CFileInfo *fileinfo, CString path, eFileTypes filetype )
{
	struct filebaseobject fbo;
	unsigned long i;

	fbo.m_eFileType  = filetype;
	fbo.m_nFileIndex = m_pUpdatingFileBaseArray->Size();
	fbo.m_nPathIndex = m_pUpdatingPathBaseArray->Size();
	fbo.m_nHashIndex = (unsigned long)-1;
	fbo.m_nSize      = fileinfo->size;
	fbo.m_tModTime   = fileinfo->st_m_time;

	i = m_pUpdatingBaseArray->Size()/sizeof(struct filebaseobject);

	m_pUpdatingBaseArray->Append( (const char*)&fbo, sizeof(struct filebaseobject) );
	m_pUpdatingFileBaseArray->Append( fileinfo->name.Data(), fileinfo->name.Length()+1 );
	m_pUpdatingPathBaseArray->Append( path.Data(), path.Length()+1 );
	
	return i;
}

/** UpdateIndex is only used during update */
void CSearchIndex::UpdateIndex( ulonglong index, struct filebaseobject * fbo )
{
	if ( (index*sizeof(struct filebaseobject)) < m_pUpdatingBaseArray->Size() )
	{
		memcpy( m_pUpdatingBaseArray->Data()+(index*sizeof(struct filebaseobject)), fbo, sizeof(struct filebaseobject) );
	}
}

/** */
bool CSearchIndex::FindHash( unsigned char * hash, ulonglong * hi )
{
	ulonglong i;
	
	for(i=*hi;i<m_pHashIndex->Size();i+=TigerTree::HASH_SIZE)
		if ( memcmp(hash,m_pHashIndex->Data()+i,TigerTree::HASH_SIZE) == 0 )
		{
			*hi = i;
			return TRUE;
		}
		
	return FALSE;
}

/** */
bool CSearchIndex::FindHashDuringUpdate( unsigned char * hash, ulonglong * hi )
{
	ulonglong i;
	
	for(i=*hi;i<m_pUpdatingHashIndex->Size();i+=TigerTree::HASH_SIZE)
		if ( memcmp(hash,m_pUpdatingHashIndex->Data()+i,TigerTree::HASH_SIZE) == 0 )
		{
			*hi = i;
			return TRUE;
		}
		
	return FALSE;
}

/** */
CString CSearchIndex::GetHash( ulonglong hbi )
{
	CString s;
	CBase32 base32;
	CByteArray dst,src;
	struct hashbaseobject * hbo;

	if ( hbi < m_pHashBaseArray->Size() )
	{
		hbo = (struct hashbaseobject *)(m_pHashBaseArray->Data()+hbi);
		src.Append(m_pHashIndex->Data()+hbo->m_nHashIndex,TigerTree::HASH_SIZE);
		base32.Encode( &dst, &src );
		if ( dst.Size() > 0 )
		{
			s.Set((const char*)dst.Data(),dst.Size());
			s = "TTH:"+s;
		}
	}
	
	return s;
}

/** */
bool CSearchIndex::HashBaseIndexFromHashIndex( ulonglong hi, ulonglong * hbi )
{
	ulonglong i;
	struct hashbaseobject * hbo;

	for(i=0;i<m_pHashBaseArray->Size();i+=sizeof(struct hashbaseobject))
	{
		hbo = (struct hashbaseobject *)(m_pHashBaseArray->Data()+i);
		if ( hbo->m_nHashIndex == hi )
		{
			*hbi = i;
			return TRUE;
		}
	}
			
	return FALSE;
}

/** */
bool CSearchIndex::HashBaseIndexFromHashIndexDuringUpdate( ulonglong hi, ulonglong * hbi )
{
	ulonglong i;
	struct hashbaseobject * hbo;

	for(i=0;i<m_pUpdatingHashBaseArray->Size();i+=sizeof(struct hashbaseobject))
	{
		hbo = (struct hashbaseobject *)(m_pUpdatingHashBaseArray->Data()+i);
		if ( hbo->m_nHashIndex == hi )
		{
			*hbi = i;
			return TRUE;
		}
	}
			
	return FALSE;
}

/** */
bool CSearchIndex::BaseIndexFromHashBaseIndex( ulonglong hbi, ulonglong * bi )
{
	ulonglong i;
	struct filebaseobject * fbo;
	
	for(i=0;i<m_pBaseArray->Size();i+=sizeof(struct filebaseobject))
	{
		fbo = (struct filebaseobject*)(m_pBaseArray->Data()+i);
		
		if ( fbo->m_nHashIndex == hbi )
		{
			*bi = i/sizeof(struct filebaseobject);
			return TRUE;
		}
	}
	
	return FALSE;
}

/** */
bool CSearchIndex::FindHashBaseIndex( struct filebaseobject * fbo, ulonglong * hbi )
{
	ulonglong i;
	struct hashbaseobject * hbo;

	for(i=0;i<m_pHashBaseArray->Size();i+=sizeof(struct hashbaseobject))
	{
		hbo = (struct hashbaseobject *)(m_pHashBaseArray->Data()+i);
		
		if ( Compare(fbo,hbo) )
		{
			*hbi = i;
			return TRUE;
		}
	}
	
	return FALSE;
}

/** */
bool CSearchIndex::FindHashBaseIndexDuringUpdate( struct filebaseobject * fbo, ulonglong * hbi )
{
	ulonglong i;
	struct hashbaseobject * hbo;

	for(i=0;i<m_pUpdatingHashBaseArray->Size();i+=sizeof(struct hashbaseobject))
	{
		hbo = (struct hashbaseobject *)(m_pUpdatingHashBaseArray->Data()+i);
		
		if ( CompareDuringUpdate(fbo,hbo) )
		{
			*hbi = i;
			return TRUE;
		}
	}
	
	return FALSE;
}

/** */
bool CSearchIndex::Compare( struct filebaseobject * fbo, struct hashbaseobject * hbo )
{
	if ( fbo->m_nSize != hbo->m_nSize )
		return FALSE;
	if ( fbo->m_tModTime != hbo->m_tModTime )
		return FALSE;

	CString s1,s2;
	
	s1 = (char*)m_pFileBaseArray->Data()+fbo->m_nFileIndex;
	s2 = (char*)m_pHashFileBaseArray->Data()+hbo->m_nFileIndex;
	
	if ( s1 != s2 )
		return FALSE;
	
	s1 = (char*)m_pPathBaseArray->Data()+fbo->m_nPathIndex;
	s2 = (char*)m_pHashPathBaseArray->Data()+hbo->m_nPathIndex;
	
	if ( s1 != s2 )
		return FALSE;

	return TRUE;
}

/** */
bool CSearchIndex::CompareDuringUpdate( struct filebaseobject * fbo, struct hashbaseobject * hbo )
{
	if ( fbo->m_nSize != hbo->m_nSize )
		return FALSE;
	if ( fbo->m_tModTime != hbo->m_tModTime )
		return FALSE;

	CString s1,s2;
	
	s1 = (char*)m_pUpdatingFileBaseArray->Data()+fbo->m_nFileIndex;
	s2 = (char*)m_pUpdatingHashFileBaseArray->Data()+hbo->m_nFileIndex;
	
	if ( s1 != s2 )
		return FALSE;
	
	s1 = (char*)m_pUpdatingPathBaseArray->Data()+fbo->m_nPathIndex;
	s2 = (char*)m_pUpdatingHashPathBaseArray->Data()+hbo->m_nPathIndex;
	
	if ( s1 != s2 )
		return FALSE;

	return TRUE;
}

/** */
void CSearchIndex::AddHashIndex( ulonglong filebaseindex, unsigned char * hash, unsigned char * leaves, unsigned int lsize )
{
	struct filebaseobject fbo;
	struct hashbaseobject hbo;
	char * c;
	ulonglong i,hi,hbi;
	
	if ( !GetFileBaseObjectDuringUpdate(filebaseindex,&fbo) )
	{
		return;
	}
	
	hi = 0;
	if ( FindHashDuringUpdate(hash,&hi) )
	{
		if ( HashBaseIndexFromHashIndexDuringUpdate(hi,&hbi) )
		{
			if ( CompareDuringUpdate(&fbo,(struct hashbaseobject *)(m_pUpdatingHashBaseArray->Data()+hbi)) )
			{
				printf("hash found\n");
				fbo.m_nHashIndex = hbi;
				UpdateIndex(filebaseindex,&fbo);
				return;
			}
		}
	}
	
	hbo.m_nSize = fbo.m_nSize;
	hbo.m_tModTime = fbo.m_tModTime;
	hbo.m_nFileIndex = m_pUpdatingHashFileBaseArray->Size();
	hbo.m_nPathIndex = m_pUpdatingHashPathBaseArray->Size();
	hbo.m_nHashIndex = m_pUpdatingHashIndex->Size();
	
	if ( (lsize == 0) || (leaves == 0) )
	{
		hbo.m_nHashLeavesIndex = (unsigned long)-1;
	}
	else
	{
		if ( m_pNewHashLeaves == 0 )
		{
			m_pNewHashLeaves = new CByteArray();
			hashleavessize = CDir().getFileSize(CConfig::Instance()->GetConfigPath()+"hashleaves.bin",FALSE);
		}
		
		i = lsize;
		hbo.m_nHashLeavesIndex = hashleavessize + m_pNewHashLeaves->Size();
		m_pNewHashLeaves->Append((unsigned char*)&i,sizeof(i));
		m_pNewHashLeaves->Append(leaves,lsize);
	}
	
	fbo.m_nHashIndex = m_pUpdatingHashBaseArray->Size();

	m_pUpdatingHashBaseArray->Append((const char*)&hbo, sizeof(struct hashbaseobject));
	c = (char*)m_pUpdatingFileBaseArray->Data()+fbo.m_nFileIndex;
	m_pUpdatingHashFileBaseArray->Append(c,strlen(c)+1);
	c = (char*)m_pUpdatingPathBaseArray->Data()+fbo.m_nPathIndex;
	m_pUpdatingHashPathBaseArray->Append(c,strlen(c)+1);
	m_pUpdatingHashIndex->Append( hash, TigerTree::HASH_SIZE );

	UpdateIndex(filebaseindex,&fbo);
}

/** */
void CSearchIndex::InitIndex()
{
	int i;

	// init array
	m_pUpdatingSearchFileIndex->SetSize(sizeof(struct long_s)*0x100);

	for(i=0;i<=0xFF;i++)
	{
		m_nUpdatingSearchFileIndex = (struct long_s*)m_pUpdatingSearchFileIndex->Data();

		m_nUpdatingSearchFileIndex[i].n = 0;

		if ( m_pSearchArray[i] != 0 )
		{
			m_nUpdatingSearchFileIndex[i].n = m_pUpdatingSearchFileIndex->Size();
			m_pUpdatingSearchFileIndex->Append(m_pSearchArray[i]->Data(),m_pSearchArray[i]->Size());
			delete m_pSearchArray[i];
		}
	}
}

/** */
CString CSearchIndex::GetFileName( ulonglong i )
{
	CString s = "";
	struct filebaseobject * fbo;

	if ( (i*sizeof(struct filebaseobject)) < m_pBaseArray->Size() )
	{
		fbo = (struct filebaseobject *) (m_pBaseArray->Data()+(i*sizeof(struct filebaseobject)));

		s = (char*) (m_pFileBaseArray->Data()+fbo->m_nFileIndex);
	}

	return s;
}

/** */
CString CSearchIndex::GetFileNameDuringUpdate( ulonglong i )
{
	CString s = "";
	struct filebaseobject * fbo;

	if ( (i*sizeof(struct filebaseobject)) < m_pUpdatingBaseArray->Size() )
	{
		fbo = (struct filebaseobject *) (m_pUpdatingBaseArray->Data()+(i*sizeof(struct filebaseobject)));

		s = (char*) (m_pUpdatingFileBaseArray->Data()+fbo->m_nFileIndex);
	}

	return s;
}

struct searchindexobject * CSearchIndex::SearchIndex( CString & s )
{
	unsigned long i,is,ie;
	unsigned int in = s.Data()[0]&0xFF;
	struct searchindexobject * psio = 0;

	if ( (!m_nSearchFileIndex) || (m_nSearchFileIndex[in].n == 0) )
	{
		return 0;
	}

	// set start
	is = m_nSearchFileIndex[in].n;

	// find end
	for(ie=0,i=in+1;i<=0xFF;i++)
	{
		if ( m_nSearchFileIndex[i].n != 0 )
		{
			ie = m_nSearchFileIndex[i].n;
			break;
		}
	}

	if ( ie == 0 )
	{
		ie = m_pSearchFileIndex->Size();
	}

	i = 0;

	while( (is+(i*sizeof(struct searchindexobject))) < ie )
	{
		psio = (struct searchindexobject *) (m_pSearchFileIndex->Data()+(is+(i*sizeof(struct searchindexobject))));

		if ( s == (char*)(m_pSearchBase->Data()+psio->m_nIndex) )
		{
			break;
		}

		i++;
		psio = 0;
	}

	return psio;
}

struct searchindexobject * CSearchIndex::FindIndex( CString & s )
{
	long i=0;
	unsigned int in = s.Data()[0]&0xFF;
	struct searchindexobject * psio = 0;

	if ( m_pSearchArray[in] == 0 )
	{
		return 0;
	}

	while( (i*sizeof(struct searchindexobject)) < m_pSearchArray[in]->Size() )
	{
		psio = (struct searchindexobject *) (m_pSearchArray[in]->Data()+(i*sizeof(struct searchindexobject)));

		if ( s == (char*)(m_pUpdatingSearchBase->Data()+psio->m_nIndex) )
		{
			break;
		}

		i++;
		psio = 0;
	}

	return psio;
}

void CSearchIndex::AddIndex( CString & s, ulonglong filebaseindex )
{
	unsigned int in;
	struct searchindexobject sio, *psio;
	struct fileindexobject fio, *pfio;

	in = s.Data()[0]&0xFF;
			
	if ( m_pSearchArray[in] == 0 )
	{
		m_pSearchArray[in] = new CByteArray();
	}
			
	if ( (psio = FindIndex(s) ) != 0 )
	{
		pfio = (struct fileindexobject*) (m_pUpdatingSearchIndex->Data()+psio->m_nFileIndex);

		while(pfio->m_nNext!=0)
		{
			pfio = (struct fileindexobject*) (m_pUpdatingSearchIndex->Data()+pfio->m_nNext);
		}

		pfio->m_nNext = m_pUpdatingSearchIndex->Size();
	}
	else
	{
		sio.m_nIndex     = m_pUpdatingSearchBase->Size();
		sio.m_nFileIndex = m_pUpdatingSearchIndex->Size();

		m_pUpdatingSearchBase->Append(s.Data(),s.Length()+1);
		m_pSearchArray[in]->Append( (const char*)&sio, sizeof(struct searchindexobject) );
	}

	fio.m_nFileBaseIndex = filebaseindex;
	fio.m_nNext          = 0;

	m_pUpdatingSearchIndex->Append( (const char*)&fio, sizeof(struct fileindexobject) );
}

/** */
CByteArray * CSearchIndex::GetHashLeaves( CString tth )
{
	CByteArray * ba = 0;
	ulonglong hi = 0, hbi = 0, lsize = 0;
	unsigned long hli = 0;
	struct hashbaseobject * hbo = 0;
	CBase32 base32;
	CByteArray dst;
	
	if (base32.Decode(&dst, &tth) != TigerTree::HASH_SIZE)
	{
		DPRINTF("GetHashLeaves: Decoded TTH length != TigerTree::HASH_SIZE\n");
		return 0;
	}
	
	if ( FindHash(dst.Data(), &hi) )
	{
		//DPRINTF("GetHashLeaves: HashIndex=%llu\n", hi);
		
		if ( HashBaseIndexFromHashIndex( hi, &hbi ) )
		{
			//DPRINTF("GetHashLeaves: HashBaseIndex=%llu\n", hbi);
			hbo = (struct hashbaseobject *)(m_pHashBaseArray->Data()+hbi);
			
			// hash leaves index
			hli = hbo->m_nHashLeavesIndex;
			
			if (hli == (unsigned long)-1)
			{
				DPRINTF("GetHashLeaves: No Leaves available.\n");
				return 0;
			}
			
			//DPRINTF("GetHashLeaves: HashLeavesIndex=%llu\n", hli);
			
			/* Old code for in memory hash leaves*/
			/* lsize = * (ulonglong *) (m_pHashLeaves->Data()+hli);
			
			//DPRINTF("GetHashLeaves: lsize=%llu\n", lsize);
			
			ba = new CByteArray();
			ba->Append( (m_pHashLeaves->Data()+(hli + sizeof(lsize))), lsize );
			
			return ba; */
			
			/* New code for on disk hash leaves */
			CFile leaffile;
			
			if ( leaffile.Open(CConfig::Instance()->GetConfigPath()+"hashleaves.bin", IO_RAW | IO_READONLY) == FALSE )
			{
				printf("GetHashLeaves: cannot open hashleaves.bin\n");
				printf("GetHashLeaves: hli=%lu tth=%s\n", hli, tth.Data());
				return 0;
			}
			
			if ( leaffile.Seek(hli, 0) == FALSE )
			{
				printf("GetHashLeaves: seek failed, hash database corrupt\n");
				printf("GetHashLeaves: hli=%lu tth=%s\n", hli, tth.Data());
				leaffile.Close();
				return 0;
			}
			
			long read1 = 0, read2 = 0;
			read1 = leaffile.Read( (char *)&lsize, sizeof(lsize) );
			if (read1 != sizeof(lsize))
			{
				printf("GetHashLeaves: error didn't read leaves size, hash database corrupt\n");
				printf("GetHashLeaves: hli=%lu tth=%s\n", hli, tth.Data());
				leaffile.Close();
				return 0;
			}
			
			//DPRINTF("GetHashLeaves: lsize=%llu\n", lsize);
			
			ba = new CByteArray( lsize );
			
			read2 = leaffile.Read( (char*)ba->Data(), lsize );
			if (read2 != lsize)
			{
				printf("GetHashLeaves: error didn't read enough data, hash database corrupt\n");
				printf("GetHashLeaves: hli=%lu tth=%s\n", hli, tth.Data());
				leaffile.Close();
				delete ba;
				return 0;
			}
			
			leaffile.Close();
			
			return ba;
		}
		else
		{
			printf("HashBaseIndexFromHashIndex( hi, &hbi ) failed for %llu\n", hi);
			return 0;
		}
	}
	else
	{
		printf("FindHash for %s failed\n", tth.Data());
		return 0;
	}
}

/** */
bool CSearchIndex::SaveNewHashLeaves()
{
	bool res;
	
	if (m_pNewHashLeaves != 0)
	{
		
		CFile file;
		res = file.Open( CConfig::Instance()->GetConfigPath()+"hashleaves.bin",
			   IO_RAW | IO_WRITEONLY | IO_APPEND | IO_CREAT,
			   MO_IRUSR | MO_IWUSR
		);
		
		if ( res == FALSE )
		{
			printf("CSearchIndex::SaveNewHashLeaves: open hashleaves.bin failed\n");
		}
		else
		{
			long wrote = file.Write( (const char*) m_pNewHashLeaves->Data(), m_pNewHashLeaves->Size() );
			
			if ( wrote == m_pNewHashLeaves->Size() )
			{
				res = TRUE;
			}
			else
			{
				printf("CSearchIndex::SaveNewHashLeaves: only wrote %ld out of %ld bytes\n",wrote,m_pNewHashLeaves->Size());
				if ( wrote > 0 )
				{
					printf("CSearchIndex::SaveNewHashLeaves: hashleaves.bin has been corrupted, do /rebuild\n");
				}
				res = FALSE;
			}
			
			file.Close();
			
			/* If data was written to hashleaves.bin, even if only some, delete m_pNewHashLeaves */
			if ( wrote > 0 )
			{
				delete m_pNewHashLeaves;
				m_pNewHashLeaves = 0;
			}
		}
	}
	else
	{
		//DPRINTF("CSearchIndex::SaveNewHashLeaves: nothing to save\n");
		res = FALSE;
	}
	
	return res;
}

/** */
void CSearchIndex::RebuildLists()
{
	CByteArray * pNewBaseArray     = new CByteArray();
	CByteArray * pNewFileBaseArray = new CByteArray();
	CByteArray * pNewPathBaseArray = new CByteArray();
	
	CByteArray * pNewHashBaseArray     = new CByteArray();
	CByteArray * pNewHashFileBaseArray = new CByteArray();
	CByteArray * pNewHashPathBaseArray = new CByteArray();
	CByteArray * pNewHashIndex         = new CByteArray();
	CByteArray * pNewHashLeaves        = 0;
	
	// we cannot do anything with the hash leaves if hashing is disabled...
	// and just in case someone changes the setting while this is running, store the value
	bool disablehashlist = CConfig::Instance()->GetDisableHashList();
	if ( disablehashlist == FALSE )
	{
		pNewHashLeaves = new CByteArray();
	}
	
	ulonglong count = IndexCount();
	
	printf("##### Before Rebuild #####\n");
	PrintDatabaseStats();
	printf("##########################\n");
	
	CFile file;
	CDir cdir;
	CBase32 b32;
	
	for (ulonglong i = 0; i < count; i++)
	{
		if ( CFileManager::Instance()->Stopped() )
		{
			printf("RebuildLists stopped\n");
			
			delete pNewBaseArray;
			delete pNewFileBaseArray;
			delete pNewPathBaseArray;
			
			delete pNewHashBaseArray;
			delete pNewHashFileBaseArray;
			delete pNewHashPathBaseArray;
			delete pNewHashIndex;
			if ( pNewHashLeaves != 0 )
			{
				delete pNewHashLeaves;
			}
			
			return;
		}
		
		struct filebaseobject fbo;
		
		if ( GetFileBaseObject( i, &fbo ) == FALSE )
		{
			DPRINTF("RebuildLists: failed to get filebaseobject for %llu\n", i);
			continue;
		}
		
		CString vpath = (char*) m_pPathBaseArray->Data()+fbo.m_nPathIndex;
		
		// top level folders have "/" as the path and their name as the filename
		CString filenamewithpath = "";
		
		if ( vpath.Length() > 1 )
		{
			filenamewithpath = vpath;
			filenamewithpath += DIRSEPARATOR;
		}
		
		CString filename = (char*) m_pFileBaseArray->Data()+fbo.m_nFileIndex;
		filenamewithpath += filename;
		
		CString realpath = CConfig::Instance()->AliasToPath(filenamewithpath);
		
		DPRINTF( "'%s' ", realpath.Data() );
		
		// check if file no longer shared
		if ( realpath == "" )
		{
			DPRINTF( "%s no longer shared, entry removed\n", filenamewithpath.Data() );
			continue;
		}
		
		// if file exists
		if ( cdir.IsDir(realpath,FALSE) == TRUE )
		{
			DPRINTF("is existing folder\n");
			fbo.m_nFileIndex = pNewFileBaseArray->Size();
			fbo.m_nPathIndex = pNewPathBaseArray->Size();
			fbo.m_nHashIndex = (unsigned long)-1;
			
			pNewBaseArray->Append( (const char*)&fbo, sizeof(struct filebaseobject) );
			pNewFileBaseArray->Append( filename.Data(), filename.Length() + 1 );
			pNewPathBaseArray->Append( vpath.Data(), vpath.Length() + 1 );
		}
		else if ( file.Open(realpath, IO_RAW | IO_READONLY) == TRUE )
		{
			file.Close();
			DPRINTF("exists ");
			
			fbo.m_nFileIndex = pNewFileBaseArray->Size();
			fbo.m_nPathIndex = pNewPathBaseArray->Size();
			
			// get TTH and hash leaves
			if ( fbo.m_nHashIndex != (unsigned long)-1 )
			{
				struct hashbaseobject * oldhbo = (struct hashbaseobject *) (m_pHashBaseArray->Data()+fbo.m_nHashIndex);

				CByteArray dst,src;
				src.Append( m_pHashIndex->Data()+oldhbo->m_nHashIndex, TigerTree::HASH_SIZE );
				b32.Encode( &dst, &src );
				CString tth;
				tth.Set( (const char*)dst.Data(), dst.Size() );
				
				DPRINTF("tth=%s\n", tth.Data() );
				
				CByteArray * hashleaves = 0;
				ulonglong lsize = 0;
				
				if ( disablehashlist == FALSE)
				{
					hashleaves = GetHashLeaves( tth );
					
					if ( hashleaves != 0 )
					{
						lsize = hashleaves->Size();
						
						if ( CFileHasher::ValidateHashLeaves( &src, hashleaves, cdir.getFileSize(realpath,FALSE) ) == FALSE )
						{
							delete hashleaves;
							hashleaves = 0;
							lsize = 0;
						}
					}
					
					if ( lsize == 0 )
					{
						printf("leaves not found or corrupt for %s, generating\n", tth.Data());
						CFileHasherThread hasher( realpath );
						hasher.Start();
						while ( hasher.GetStatus() == efhsWorking )
						{
							if ( CFileManager::Instance()->Stopped() )
							{
								printf("RebuildLists: CFileManager stopped, stopping hasher\n");
								hasher.StopHashing();
								
								delete pNewBaseArray;
								delete pNewFileBaseArray;
								delete pNewPathBaseArray;
								
								delete pNewHashBaseArray;
								delete pNewHashFileBaseArray;
								delete pNewHashPathBaseArray;
								delete pNewHashIndex;
								if ( pNewHashLeaves != 0 )
								{
									delete pNewHashLeaves;
								}
								
								return;
							}
#ifdef WIN32
							Sleep(100);
#else
							usleep(100000);
#endif
						}
				
						if ( hasher.GetStatus() == efhsFinished )
						{
							if ( tth != hasher.GetHashRoot() )
							{
								printf("RebuildLists: tth root mismatch, old=%s new=%s ,fixing ", tth.Data(), hasher.GetHashRoot().Data());
								src.SetSize(0);
								src.Append( hasher.HashRootDirect()->Data(), hasher.HashRootDirect()->Size() );
							}
						
							if ( hashleaves == 0 )
							{
								hashleaves = new CByteArray();
							}
							
							hashleaves->SetSize(0);
							lsize = hasher.HashLeavesDirect()->Size();
							//printf("new lsize=%llu new hli=%lu (unsigned long)-1=%lu\n", lsize, pNewHashLeaves->Size(), (unsigned long)-1 );
							hashleaves->Append( hasher.HashLeavesDirect()->Data(), lsize );
						}
						else
						{
							printf("RebuildLists: problem generating hash data ");
							lsize = 0;
							hashleaves->SetSize(0);
						}
					}
				}
				
				fbo.m_nHashIndex = pNewHashBaseArray->Size();
				
				struct hashbaseobject hbo;
				hbo.m_nSize    = fbo.m_nSize;
				hbo.m_tModTime = fbo.m_tModTime;

				hbo.m_nPathIndex = pNewHashPathBaseArray->Size();
				hbo.m_nFileIndex = pNewHashFileBaseArray->Size();
				hbo.m_nHashIndex = pNewHashIndex->Size();
				
				if ( disablehashlist == FALSE )
				{
					if ( lsize > 0 )
					{
						hbo.m_nHashLeavesIndex = pNewHashLeaves->Size();
					}
					else
					{
						hbo.m_nHashLeavesIndex = (unsigned long)-1;
					}
					
				}
				else
				{
					hbo.m_nHashLeavesIndex = oldhbo->m_nHashLeavesIndex;
				}
				
				pNewHashBaseArray->Append( (const char*)&hbo, sizeof(struct hashbaseobject) );
				pNewHashPathBaseArray->Append( vpath.Data(), vpath.Length() + 1 );
				pNewHashFileBaseArray->Append( filename.Data(), filename.Length() + 1 );
				pNewHashIndex->Append( src.Data(), TigerTree::HASH_SIZE );
				
				if ( lsize > 0 )
				{
					pNewHashLeaves->Append( (unsigned char*)&lsize, sizeof(lsize) );
					pNewHashLeaves->Append( hashleaves->Data(), lsize );
				}
				
				if ( hashleaves != 0 )
				{
					delete hashleaves;
				}
			}
			else
			{
				DPRINTF("no tth\n");
			}
			
			pNewBaseArray->Append( (const char*)&fbo, sizeof(struct filebaseobject) );
			pNewFileBaseArray->Append( filename.Data(), filename.Length() + 1 );
			pNewPathBaseArray->Append( vpath.Data(), vpath.Length() + 1 );
		}
		else
		{
			DPRINTF("does not exist, entry removed\n");
		}
	}
	
	// delete old lists
	delete m_pBaseArray;
	delete m_pFileBaseArray;
	delete m_pPathBaseArray;
	
	delete m_pHashBaseArray;
	delete m_pHashFileBaseArray;
	delete m_pHashPathBaseArray;
	delete m_pHashIndex;
	
	// replace old lists with new lists
	m_pBaseArray     = pNewBaseArray;
	m_pFileBaseArray = pNewFileBaseArray;
	m_pPathBaseArray = pNewPathBaseArray;
	
	m_pHashBaseArray     = pNewHashBaseArray;
	m_pHashFileBaseArray = pNewHashFileBaseArray;
	m_pHashPathBaseArray = pNewHashPathBaseArray;
	m_pHashIndex         = pNewHashIndex;
	
	if ( disablehashlist == FALSE )
	{
		bool res = pNewHashLeaves->SaveToFile(CConfig::Instance()->GetConfigPath()+"hashleaves.bin");
		if ( res == FALSE )
		{
			/*
			 * most likely out of disk space, a permissions problem
			 * or filesystem remounted read only due to kernel detecting an error
			 */
			printf("Failed to save rebuilt hash leaf data, hash leaf data destroyed\n");
		}
		delete pNewHashLeaves;
	}
	
	printf("##### After Rebuild #####\n");
	PrintDatabaseStats();
	printf("#########################\n");
}

/** */
void CSearchIndex::PrintDatabaseStats()
{
	printf("BaseArray size=%lu itemsize=%d items=%lu \n", m_pBaseArray->Size(), sizeof(struct filebaseobject), m_pBaseArray->Size()/sizeof(struct filebaseobject) );
	printf("FileBaseArray size=%lu\n", m_pFileBaseArray->Size());
	printf("PathBaseArray size=%lu\n", m_pPathBaseArray->Size());
	
	printf("SearchIndex size=%lu itemsize=%d items=%lu\n", m_pSearchIndex->Size(), sizeof(struct fileindexobject), m_pSearchIndex->Size()/sizeof(struct fileindexobject));
	printf("SearchFileIndex size=%lu\n", m_pSearchFileIndex->Size());
	printf("SearchBase size=%lu\n",m_pSearchBase->Size());
	
	printf("HashBaseArray size=%lu itemsize=%d items=%lu\n", m_pHashBaseArray->Size(), sizeof(struct hashbaseobject), m_pHashBaseArray->Size()/sizeof(struct hashbaseobject));
	printf("HashFileBaseArray size=%lu\n", m_pHashFileBaseArray->Size());
	printf("HashPathBaseArray size=%lu\n", m_pHashPathBaseArray->Size());
	printf("HashIndex size=%lu itemsize=%d items=%lu\n", m_pHashIndex->Size(), TigerTree::HASH_SIZE, m_pHashIndex->Size()/TigerTree::HASH_SIZE);
	
	printf("HashLeaves size=%lu (on disk)\n", CDir().getFileSize(CConfig::Instance()->GetConfigPath()+"hashleaves.bin",FALSE));
}
