/***************************************************************************
                          ctransfer.cpp  -  description
                             -------------------
    begin                : Fri Feb 22 2002
    copyright            : (C) 2002-2003 by Mathias Kster
    email                : mathen@users.berlios.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#ifndef WIN32
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#else
#include <io.h>
#endif

#include "dcos.h"
#include "dclib.h"
#include "dcobject.h"
#include "cmessagehandler.h"
#include "cencrypt.h"
#include "core/cdir.h"
#include "core/cbytearray.h"
#include "cconfig.h"
#include "core/platform.h"
#include "cfilemanager.h"

#include "core/clogfile.h"

#include "ctransfer.h"

CTransfer::CTransfer( bool listener )
{
	pCallback = 0;

	pByteArray = new CByteArray();

	m_bListener   = listener;
	sNick         = "";
	sSrcFilename  = "";
	sDstFilename  = "";
	sBuffer       = "";
	eMedium       = eltNONE;
	eSrcDirection = edNONE;
	eDstDirection = edNONE;
	eMode         = estNONE;

	m_nStartPosition = 0;
	m_nEndPosition   = 0;
	m_nSrcLevel      = 0;
	m_nDstLevel      = 0;
	starttime.tv_sec = 0;
	starttime.tv_usec= 0;
	m_nFileSize      = 0;
	m_nTransfered    = 0;
	m_nTransferRate  = 0;
	m_nChunkSize     = 0;

	sDstNick = "";

	sHubName = "";
	sHubHost = "";

	m_nTransferID = 0;

	m_bIdle = TRUE;
	m_eTransferState = etsNONE;

	m_eTransferType = ettNONE;

	// init transferrate
	memset(&m_AverageTransferRate,0,sizeof(m_AverageTransferRate));
	m_AverageTransferRate.index=0;
	gettimeofday(&m_AverageTransferRate.tv[0],NULL);
	
	if ( CConfig::Instance() )
	{
		m_nReadFileBufferSize = CConfig::Instance()->GetReadFileBufferSize();
	}
	else
	{
		m_nReadFileBufferSize = 40960;
	}
	
	m_nZlibStatus = 0;
}

CTransfer::~CTransfer()
{
	SetCallBackFunction(0);

	TransferMutex.Lock();

	if (pByteArray)
	{
		delete pByteArray;
		pByteArray = 0;
	}

	m_File.Close();

	TransferMutex.UnLock();
	
	if ( m_nZlibStatus == 1 ) // -1 error status will have already been reported elsewhere
	{
		DPRINTF("CTransfer::~CTransfer: zlib not finished\n");
	}
}

/** */
int CTransfer::CallBack_SendObject( CObject * Object )
{
	int err;

	if ( pCallback != 0 )
	{
		err = pCallback->notify( this, Object );
	}
	else
	{
		err = DC_CallBack( Object );
	}

	if ( err == -1 )
	{
		DPRINTF("CallBack failed (state)...\n");
		delete Object;
	}

	return err;
}

/** */
int CTransfer::CallBack_SendError( CString msg )
{
	CMessageError * Object = new CMessageError();

	Object->m_sError = msg;

	return CallBack_SendObject(Object);
}

/** */
void CTransfer::SetBuffer( CByteArray *ba )
{
	TransferMutex.Lock();
	pByteArray->Append(ba->Data(),ba->Size());
	TransferMutex.UnLock();
}

/** */
long CTransfer::GetBuffer( CByteArray *ba )
{
	TransferMutex.Lock();
	long i;
	ba->Append(pByteArray->Data(),pByteArray->Size());
	i = pByteArray->Size();
	TransferMutex.UnLock();
	return i;
}

/** */
void CTransfer::InitTime()
{
	TransferMutex.Lock();
	gettimeofday(&starttime,NULL);
	TransferMutex.UnLock();
}

/** */
ulonglong CTransfer::GetTransferrate()
{
	TransferMutex.Lock();

	ulonglong r;

	if ( (starttime.tv_sec == 0) || (m_bIdle == TRUE) )
	{
		TransferMutex.UnLock();
		return 0;
	}

	r = GetTraffic();

	TransferMutex.UnLock();

	return r;
}

/** */
ulonglong CTransfer::GetBytesForTransferrate( ulonglong rate )
{
	TransferMutex.Lock();

	ulonglong r = 0;

	if ( starttime.tv_sec == 0 )
	{
		TransferMutex.UnLock();
		return 0;
	}

	r = GetTraffic();

	if ( r < rate )
	{
		r = rate;
	}
	else if ( r > rate )
	{
		r = 0;
	}

	TransferMutex.UnLock();

	return r;
}

/** */
void CTransfer::AddTraffic( long n )
{
	int i;
	ulonglong t1,t2;
	struct timeval tp;

	m_nTransfered += n;

	gettimeofday(&tp,NULL);

	t1 = ((tp.tv_sec*1000)+(tp.tv_usec/1000));

	i = m_AverageTransferRate.index;

	t2 = ((m_AverageTransferRate.tv[i].tv_sec*1000)+(m_AverageTransferRate.tv[i].tv_usec/1000));

	if ( (t1-t2) > 1000 )
	{
		i++;

		if ( i >= 10 )
		{
			i = 0;
		}

		m_AverageTransferRate.index = i;
		m_AverageTransferRate.tv[i] = tp;
		m_AverageTransferRate.transfered[i] = 0;
	}

	m_AverageTransferRate.transfered[i] += n;
}

/** */
ulonglong CTransfer::GetTraffic()
{
	int i;
	struct timeval tp;
	ulonglong t1,t2;
	ulonglong n = 0;

	gettimeofday(&tp,NULL);

	t1 = ((tp.tv_sec*1000)+(tp.tv_usec/1000));

	i = m_AverageTransferRate.index;

	if ( i == 9 )
		i = 0;
	else
	{
		i++;
	}

	t2 = ((m_AverageTransferRate.tv[i].tv_sec*1000)+(m_AverageTransferRate.tv[i].tv_usec/1000));

	if ( (t2 == 0) && (i > 0) )
	{
		i = 0;
	}

	t2 = ((m_AverageTransferRate.tv[i].tv_sec*1000)+(m_AverageTransferRate.tv[i].tv_usec/1000));

	if ( t2 != 0 )
	{
		t1 = t1-t2;

		if ( t1 > 0 )
		{
		for(i=0,t2=0;i<10;i++)
		{
			n += m_AverageTransferRate.transfered[i];
		}

		n = (n*1000)/t1;
		}
	}
	else
	{
		DPRINTF("time is null\n"); fflush(stdout);
	}

	return n;
}

/** */
void CTransfer::ConnectionState( eConnectionState state )
{
	CMessageConnectionState *Object;

	Object = new CMessageConnectionState();

	Object->m_eState   = state;
	Object->m_sMessage = GetSocketError();

	if ( state == estCONNECTED )
	{
		eMode = estTRANSFERHANDSHAKE;

		if ( sNick != "" )
		{
			SendMyNick( sNick );
		}
	}
	else if ( state == estSSLCONNECTED )
	{
		DPRINTF("change to ssl mode success\n");

		CallBack_SendObject(new CMessageKey());
	}
	else if ( state == estDISCONNECTED )
	{
		eMode = estNONE;
	}

	CallBack_SendObject(Object);
}

/** */
int CTransfer::StartDownload( CString dstfile, ulonglong startposition, ulonglong endposition, ulonglong size, ulonglong chunksize, CString srcfile, CString adcTTH )
{
	if ( (srcfile == "" ) && (eMedium == eltFILE) )
	{
		DPRINTF("ctransfer: wrong mode (empty file) %d\n",eMedium);
		return -1;
	}

	if ( m_bIdle == FALSE )
	{
		DPRINTF("ctransfer: other transfer is running\n");
		return -1;
	}

	if ( eMode == estTRANSFERUPLOAD )
	{
		DPRINTF("ctransfer: wrong transfer mode\n");
		return -1;
	}

	if (adcTTH.Left(4).ToUpper() == "TTH:")
	{
		printf("CTransfer::StartDownload Warning! Removed TTH: prefix from TTH\n");
		adcTTH = adcTTH.Mid(4, adcTTH.Length() - 4);
	}
	
	TransferMutex.Lock();
	eMode            = estTRANSFERDOWNLOAD;
	m_nStartPosition = startposition;
	m_nEndPosition   = endposition;
	m_nFileSize      = size;
	sDstFilename     = dstfile;
	sSrcFilename     = srcfile;
	m_nTransfered    = 0;
	m_nChunkSize     = chunksize;
	m_sTTH           = adcTTH;
	TransferMutex.UnLock();

	/** set starttime */
	InitTime();

	if ( eMedium == eltCLIENTVERSION )
	{
		m_bIdle = TRUE;
	}
	else
	{
		if ( (GetDstFilename() == DC_USER_FILELIST_HE3) && (m_MessageSupports.m_bXMLBZList == TRUE) )
		{
			if (m_MessageSupports.m_bADCGet == TRUE)
			{
				SetSrcFilename(DC_USER_FILELIST_XMLBZ);
				SendADCGet( "file", DC_USER_FILELIST_XMLBZ, 0, -1 );
			}
			else
			{
				// SendUGetBlock causes an infinite loop of filelist downloads! not anymore!
				SetSrcFilename(DC_USER_FILELIST_XMLBZ);
				SendUGetBlock( DC_USER_FILELIST_XMLBZ, startposition );
				// SendGet( DC_USER_FILELIST_XMLBZ, startposition+1 );
			}
		}
		else if ( (GetDstFilename() == DC_USER_FILELIST_HE3) && (m_MessageSupports.m_bBZList == TRUE) )
		{
			SetSrcFilename(DC_USER_FILELIST_BZ);
			SendGet( DC_USER_FILELIST_BZ, startposition+1 );
		}
		else
		{
			// if we don't have a TTH, we cannot do ADCGet
			// we'll try to fallback to other get commands, even though these may not work with DC++ >= 0.696
			if ( (m_MessageSupports.m_bADCGet == TRUE) && (m_MessageSupports.m_bTTHF == TRUE) && (adcTTH != "") )
			{
				if ( (m_MessageSupports.m_bZLIG) && (CConfig::Instance()->GetCompressedTransfers() == TRUE) )
				{
					SendADCGet( "file", "TTH/" + adcTTH, startposition, chunksize, TRUE );
					m_nDataType = 1;
				}
				else
				{
					SendADCGet ( "file", "TTH/" + adcTTH, startposition, chunksize, FALSE );
					m_nDataType = 0;
				}
			}
			else if ( chunksize == 0 )
			{
				if ( GetDstFilename() == DC_USER_FILELIST_HE3 )
				{
					SetSrcFilename(DC_USER_FILELIST_HE3);
				}
				SendGet( GetDstFilename(), startposition+1 );
				m_nDataType = 0;
			}
			else if (m_MessageSupports.m_bXMLBZList == TRUE)
			{
				if ( (m_MessageSupports.m_bZBlock == TRUE) &&
				     (CConfig::Instance()->GetCompressedTransfers() == TRUE) )
				{
					SendUGetZBlock( GetDstFilename(), startposition, chunksize );
					m_nDataType = 1;
				}
				else
				{
					SendUGetBlock( GetDstFilename(), startposition, chunksize);
					m_nDataType = 0;
				}
			}
			else if ( (m_MessageSupports.m_bChunk == FALSE) &&
			          ( (m_MessageSupports.m_bZBlock == FALSE) ||
				    (CConfig::Instance()->GetCompressedTransfers() == FALSE) ) )
			{
				SendGet( GetDstFilename(), startposition+1 );
				m_nDataType = 0;
			}
			else if ( m_MessageSupports.m_bChunk == TRUE )
			{
				SendGet( GetDstFilename(), startposition+1, chunksize );
				m_nDataType = 0;
			}
			else if ( (m_MessageSupports.m_bZBlock == TRUE) &&
			          (CConfig::Instance()->GetCompressedTransfers() == TRUE) )
			{
				SendGetZBlock( GetDstFilename(), startposition, chunksize );
				m_nDataType = 1;
			}
			else
			{
				return -1;
			}
		}
	}

	return 0;
}

/** */
int CTransfer::StartUpload( CString dstfile, ulonglong length, ulonglong pos, ulonglong chunksize, CString srcfile, bool uget, bool adcget, CString adcTTH, bool compress )
{
	if ( (srcfile == "" ) && (eMedium != eltBUFFER) )
	{
		DPRINTF("ctransfer: wrong mode %d\n",eMedium);
		return -1;
	}

	if ( m_bIdle == FALSE )
	{
		DPRINTF("ctransfer: other transfer is running\n");
		return -1;
	}

	if ( eMode == estTRANSFERDOWNLOAD )
	{
		DPRINTF("ctransfer: wrong transfer mode\n");
		return -1;
	}

	TransferMutex.Lock();
	eMode            = estTRANSFERUPLOAD;
	m_nStartPosition = pos;
	m_nEndPosition   = length;
	m_nFileSize      = length;
	sDstFilename     = dstfile;
	sSrcFilename     = srcfile;
	m_sTTH           = adcTTH;
	TransferMutex.UnLock();
	
	m_nTransfered = 0;
	// if chunksize 0 we use the len
	if ( chunksize == 0 )
		m_nChunkSize = length-pos;
	else
		m_nChunkSize = chunksize;
	m_nPendingSendBytes = 0;

	/** set starttime */
	InitTime();
	
	if ( compress )
	{
		m_nDataType = 2;
	}
	else
	{
		m_nDataType = 0;
	}
	
	if (adcget)
	{
		if (GetMedium() == eltTTHL)
		{
			int treesize = prepareTTHLsend( adcTTH );
			if (treesize == 0)
			{
				return -1;
			}
			else
			{
				SendADCSnd( "tthl", "TTH/" + adcTTH, 0, treesize, compress );
			}
		}
		else if ( (adcTTH == DC_USER_FILELIST_XMLBZ) || (adcTTH == DC_USER_FILELIST_XML) )
		{
			// damn hacks for correct response to filelist request
			SendADCSnd( "file", adcTTH, pos, m_nChunkSize, compress );
		}
		else if (adcTTH == "") // ADCGet without TTH
		{
			// escape spaces
			dstfile = dstfile.Replace(" ", "\\ ");
			
			SendADCSnd( "file", "/" + dstfile, pos, m_nChunkSize, compress );
		}
		else
		{
			SendADCSnd( "file", "TTH/" + adcTTH, pos, m_nChunkSize, compress );
		}
		
		if ( (eMode != estTRANSFERUPLOAD) || (m_bIdle == FALSE) )
		{
			DPRINTF("Warning: not in uploadmode/idle ADCSND [%d/%d]\n",eMode,m_bIdle);
		}
		else
		{
			DoInitUpload();
		}
	}
	else if ( (uget) || (compress) )
	{
		SendSending(m_nChunkSize);
		
		// don't know why this does not work, instead code out has been copied out of HandleMessage()
		//HandleMessage("$Send|", length);
		
		if ( (eMode != estTRANSFERUPLOAD) || (m_bIdle == FALSE) )
		{
			DPRINTF("Warning: not in uploadmode/idle SEND [%d/%d]\n",eMode,m_bIdle);
		}
		else
		{
			DoInitUpload();
		}
	}
	else
	{
		SendFileLength(length);
	}

	return 0;
}

/** */
bool CTransfer::DoInitDownload()
{
	CDir dir;
	bool b = FALSE;
	
	if ( eMedium == eltFILE )
	{
		int mode = 0;

		m_File.Close();

		dir.SetPath("");

		if ( dir.IsFile(sSrcFilename) == FALSE )
		{
			// create file
			mode |= IO_CREAT;
		}

		mode |= IO_RAW | IO_WRITEONLY;

		if ( m_File.Open( sSrcFilename, mode, MO_IRUSR|MO_IWUSR|MO_IRGRP|MO_IWGRP|MO_IROTH|MO_IWOTH ) == FALSE )
		{
			CallBack_SendError(strerror(errno));
			perror("File open");
		}
		else
		{
			if ( CConfig::Instance()->GetCreateFile() == TRUE )
			{
				if ( ((mode & IO_CREAT) == IO_CREAT) &&
				     (m_nFileSize > 0) )
				{
					if ( m_File.Seek(m_nFileSize-1,SEEK_SET) == TRUE )
					{
						if ( m_File.Write("\0",1) == 1 )
						{
							b = TRUE;
						}
					}
				}
			}
			else
			{
				b = TRUE;
			}
				
			if ( b == TRUE )
			{
				// seek to the startposition
				if ( m_File.Seek(m_nStartPosition,SEEK_SET) == FALSE )
				{
					b = FALSE;
				}
				else
				{
					b = TRUE;
				}
			}
				
			if ( b == FALSE )
			{
				CallBack_SendError(strerror(errno));
				perror("File seek");
				m_File.Close();
			}
		}
	}
	else
	{
		pByteArray->SetSize(0);
		b = TRUE;
	}

	return b;
}

/** */
int CTransfer::HandleMessage( char * c, int )
{
	int p=0;
	eDCMessage type;
	CObject * Object;

	CMessageHandler MessageHandler;

	CString s;

	s = c;

	while( (m_bIdle || (eMode == estTRANSFERUPLOAD)) &&
	       ((type=MessageHandler.Parse(&s,p,Object=0)) != DC_MESSAGE_PARSE_ERROR) )
	{
//		DPRINTF("transfer: handle message: %d\n",type);

		if ( !Object )
		{
			continue;
		}

		switch (type)
		{
			case DC_MESSAGE_SENDING:
			{
				CMessageSending * msg = (CMessageSending*)Object;
				
				if ( eMode != estTRANSFERDOWNLOAD || (m_bIdle == FALSE) )
				{
					DPRINTF("Warning: not in downloadmode/idle SENDING\n");

					delete Object;
					Object = 0;
				}
				else
				{
					m_nTransfered = 0;

					// update endposition
					if ( m_nEndPosition == 0 )
					{
						m_nEndPosition = msg->m_nLength;
					}

					// fix chunksize
					if ( m_nChunkSize != msg->m_nLength )
					{
						m_nChunkSize = msg->m_nLength;
					}

					// fixme why was this commented out?
					if (m_nFileSize == 0)
					{
						m_nFileSize = msg->m_nLength;
					}

					// start transfer
					if ( DoInitDownload() == TRUE )
					{
						m_bIdle = FALSE;
					}
				}

				printf("recv sending %lld\n",msg->m_nLength);fflush(stdout);
				break;
			}
			
			case DC_MESSAGE_ADCSND:
			{
				CMessageADCSnd * msg = (CMessageADCSnd*)Object;
				if ( eMode != estTRANSFERDOWNLOAD || (m_bIdle == FALSE) )
				{
					DPRINTF("Warning: not in downloadmode/idle ADCSND\n");

					delete Object;
					Object = 0;
				}
				else
				{
					m_nTransfered = 0;
					
					if ( m_nEndPosition == 0 )
					{
						m_nEndPosition = msg->m_nSize;
					}
					
					if ( m_nChunkSize != msg->m_nSize )
					{
						if ( m_nChunkSize == 0 )
						{
							m_nChunkSize = msg->m_nSize;
						}
						else
						{
							DPRINTF( "Error: wrong length: Requested=%llu Got=%lld\n", m_nChunkSize, msg->m_nSize );
							CallBack_SendError("Remote client sent wrong data - incorrect length!");
							Disconnect();
							
							delete Object;
							Object = 0;
							
							break;
						}
					}
					
					if ( m_nStartPosition != msg->m_nPos )
					{
						DPRINTF( "Error: wrong start position: Requested=%llu Got=%llu\n", m_nStartPosition, msg->m_nPos );
						CallBack_SendError("Remote client sent wrong data - incorrect start position!");
						Disconnect();
						
						delete Object;
						Object = 0;
						
						break;
					}
					
					if (m_nFileSize == 0)
					{
						m_nFileSize = msg->m_nSize;
					}
					
					// not sure if this is really necessary
					if (msg->m_bZlib == TRUE)
					{
						m_nDataType = 1;
					}
					else
					{
						m_nDataType = 0;
					}
					
					if ( DoInitDownload() == TRUE )
					{
						m_bIdle = FALSE;
					}
				}
				
				break;
			}
			case DC_MESSAGE_FILELENGTH:
			case DC_MESSAGE_LISTLEN:
			{
				CMessageFileLength * msg = (CMessageFileLength*)Object;

				if ( eMode != estTRANSFERDOWNLOAD || (m_bIdle == FALSE) )
				{
					DPRINTF("Warning: not in downloadmode/idle FILELENGTH/LISTLEN\n");

					delete Object;
					Object = 0;
				}
				else
				{
					m_nTransfered = 0;

					// update endposition
					if ( (m_nEndPosition == m_nFileSize) || (m_nEndPosition == 0) )
					{
						m_nEndPosition = msg->m_nFileLength;
					}

					// fix chunksize
					if ( (m_nChunkSize == m_nFileSize) || (m_nChunkSize == 0) )
					{
						m_nChunkSize = msg->m_nFileLength;
					}

					m_nFileSize = msg->m_nFileLength;

					// start transfer
					if ( DoInitDownload() == TRUE )
					{
						m_bIdle = FALSE;

						SendSend();
					}
				}

				break;
			}

			case DC_MESSAGE_SEND:
			{
				if ( (eMode != estTRANSFERUPLOAD) || (m_bIdle == FALSE) )
				{
					DPRINTF("Warning: not in uploadmode/idle SEND [%d/%d]\n",eMode,m_bIdle);

					delete Object;
					Object = 0;
				}
				else
				{
					DoInitUpload();
				}

				break;
			}

			case DC_MESSAGE_ERROR:
			{
				break;
			}

			case DC_MESSAGE_DIRECTION:
			{
				CMessageDirection * msg = (CMessageDirection*)Object;

				if ( eDstDirection != edNONE )
				{
					// remote will change the direction ... hoho ;-)
					if ( eDstDirection != msg->m_eDirection )
					{
						Disconnect(TRUE);
						DPRINTF("Warning: remote will change his direction (hack) ...\n");
						delete Object;
						Object = 0;
					}
				}
				else
				{
					eDstDirection = msg->m_eDirection;
					m_nDstLevel   = msg->m_nLevel;
				}

				break;
			}

			case DC_MESSAGE_MYNICK:
			{
				CMessageMyNick * msg = (CMessageMyNick*)Object;

				if ( eMode == estTRANSFERHANDSHAKE )
				{
					sDstNick = msg->m_sNick;
										
					m_nSrcLevel = rand()%0x7FFF;
				}
				else
				{
					DPRINTF("Warning: Nick msg from '%s' only allowed in handshake",msg->m_sNick.Data());

					delete Object;
					Object = 0;
				}

				break;
			}

			case DC_MESSAGE_LOCK:
			{
				CString s;
				CEncrypt Encrypt;
				CMessageLock * msg = (CMessageLock*)Object;

				// save lock message
				m_MessageLock = *msg;

				if ( eSrcDirection != edNONE )
				{
					// send supports command
					if ( msg->m_bExtProtocol == TRUE )
					{
						s = "";
						if ( CFileManager::Instance()->GetShareBufferSize(esbtBZ) > 0 )
							s += "BZList ";
						if ( (CFileManager::Instance()->GetShareBufferSize(esbtXMLBZ) > 0) && ((CConfig::Instance()->GetDisableHashList() == FALSE) || (CConfig::Instance()->GetDisableXMLListWithoutTTH() == FALSE)) )
						{
							s += "XmlBZList ";
						}
#ifdef HAVE_SSL
						CDir dir;
						if ( dir.IsFile(CConfig::Instance()->GetTransferCert(),FALSE) &&
						     dir.IsFile(CConfig::Instance()->GetTransferKey(),FALSE) )
						{
							s += "SSL ";
						}
#endif
						// add chunk support and minislots (files < 64KiB)
						s += "CHUNK MiniSlots ";
						
						// ADCGet needed for compatibility with DC++ >= 0.696
						if ( (CConfig::Instance()->GetDisableHashList() == FALSE) || (CConfig::Instance()->GetDisableADCGetWithoutTTH() == FALSE) )
						{
							s += "ADCGet ";
						}
						// TTHF needed for ADCGet
						if ( CConfig::Instance()->GetDisableHashList() == FALSE )
						{
							// TTHL support also needed for compatibility with DC++
							s += "TTHF TTHL ";
						}

						if ( CConfig::Instance()->GetCompressedTransfers() == TRUE )
						{
							s += "GetZBlock ZLIG ";
						}

						if ( s != "" )
						{
							SendSupports(s);
						}
					}

					SendDirection( eSrcDirection, m_nSrcLevel );

					s = "";
					Encrypt.Encrypt(msg->m_sData,s);

					SendKey(s);
				}
				else
				{
					Disconnect(TRUE);
				}

				break;
			}

			case DC_MESSAGE_SUPPORTS:
			{
				CMessageSupports * msg = (CMessageSupports*)Object;
								
				DPRINTF("LOCK %d %d %d\n",
				m_MessageLock.m_eClientVersion,
				m_MessageLock.m_nVersionMajor,
				m_MessageLock.m_nVersionMinor);
				
				// now we need to check some broken clients and limit support
				// warning! 0.68 has minor version 68, not 680! - 
				if ( (m_MessageLock.m_eClientVersion == eucvDCPP) &&
				     (m_MessageLock.m_nVersionMajor == 0) &&
				     ( (m_MessageLock.m_nVersionMinor >= 307) &&
				       (m_MessageLock.m_nVersionMinor <= 403) ) )
				{
					// disable xml list for dc++ 0.401
					printf("Warning! Disabling XML filelist support for DC++ between 0.307 and 0.403 inclusive.\n");
					msg->m_bXMLBZList = FALSE;
				}
				
				m_MessageSupports = *msg;

				break;
			}

			case DC_MESSAGE_KEY:
			{
#ifdef HAVE_SSL
				// start ssl socket
				if ( (m_MessageSupports.m_bSSL == TRUE) &&
				     (CConfig::Instance()->GetTransferCert() != "") &&
				     (CConfig::Instance()->GetTransferKey() != "") )
				{
					bool b;
					if ( m_bListener == TRUE )
						b = ChangeSocketMode(esmSSLSERVER, CConfig::Instance()->GetTransferCert(), CConfig::Instance()->GetTransferKey() );
					else
						b = ChangeSocketMode(esmSSLCLIENT);
					DPRINTF("change to ssl mode %d\n",b);
					delete Object;
					Object = 0;
				}
#endif

				break;
			}

			case DC_MESSAGE_MAXEDOUT:
			{
				break;
			}

			case DC_MESSAGE_CANCEL:
			{
				if ( (eMode == estTRANSFERUPLOAD) && (m_bIdle == FALSE) )
				{
					if ( eMedium == eltFILE )
					{
						m_File.Close();
					}

					// free buffer
					pByteArray->SetSize(0);

					// stop transfer
					m_bIdle = TRUE;
				}
				else
				{
					DPRINTF("Warning: Cancel msg not in uploadmode/idle SEND [%d/%d]\n",eMode,m_bIdle);

					delete Object;
					Object = 0;
				}

				break;
			}

			default:
			{
				//DPRINTF("ctransfer: unknown message %d\n",type);
			}
		}

		if (Object)
		{
			((CDCMessage*)Object)->m_eType = type;

			CallBack_SendObject(Object);
		}
	}

	return p;
}

/** */
bool CTransfer::GetEncrypted()
{
	bool res = FALSE;

#ifdef HAVE_SSL
	if ( (GetSocketMode() == esmSSLCLIENT) ||
	     (GetSocketMode() == esmSSLSERVER) )
	{
		res = TRUE;
	}
#endif

	return res;
}

/** */
void CTransfer::DataSend()
{
	int l = 0;
	ulonglong len=0;
	ulonglong sendbyte;
	ulonglong freebuf;
	CMessageTransfer *Object;
	int lo;

	if ( (eMode == estTRANSFERUPLOAD) && (m_bIdle == FALSE) )
	{
		for(lo=0;(lo<5) && (!m_bIdle);lo++)
		{

		sendbyte = 0;

		// check send rate
		if ( m_nTransferRate != 0 )
		{
			sendbyte = GetBytesForTransferrate(m_nTransferRate);

			if ( sendbyte == 0 )
			{
				break;
			}
		}

		freebuf = m_nReadFileBufferSize; //GetFreeSendBufferSize();
		
		// should we test for zblock also?
		if ( ((m_MessageSupports.m_bChunk == FALSE) && (m_MessageSupports.m_bADCGet == FALSE) && (m_MessageSupports.m_bXMLBZList == FALSE) && (m_MessageSupports.m_bZBlock == FALSE)) || (m_nChunkSize == 0) )
		{
			len = m_nFileSize - m_nStartPosition - m_nTransfered;
		}
		else
		{
			len = m_nChunkSize - m_nTransfered;
		}

		if ( eMedium == eltFILE )
		{
			//DPRINTF("%llu %llu %llu %llu\n",len,lStartPosition,lChunkSize,lCurrentPosition);

			// load data from filehandle and send ...
			if ( m_nFileBufferPos == m_nFileBufferSize )
			{
				if ( len > m_nReadFileBufferSize )
				{
					len = m_nReadFileBufferSize;
				}
				
				if ( m_nDataType == 2 )
				{
					pByteArray->SetSize(0);
					m_nFileBufferSize = 0;
					
					CByteArray readbuffer;
					
					long fileBytesRead = 0;
					
					while ( pByteArray->Size() < (len / 2) )
					{
						bool more = TRUE;
						int inlen = 0;
						int outlen = 0;
						CByteArray temp;
						
						if ( m_nTransfered + fileBytesRead >= m_nChunkSize )
						{
							//printf("No more data ");
							more = FALSE;
							
							// this just causes a world of pain
							/* res = TRUE;
							
							while ( res )
							{
								temp.SetSize( m_nReadFileBufferSize );
								inlen = 0;
								outlen = temp.Size();
								
								res = m_Deflater.DeflateBlock( (char *)0, &inlen, (char *)temp.Data(), &outlen, more );
								
								if ( outlen > 0 );
								{
									printf(" got %d ", outlen);
									pByteArray->Append( temp.Data(), outlen );
									m_nFileBufferSize += outlen;
								}
							} */
							
							//printf("DeflateEnd\n");
							break;
						}
						else
						{
							// limit to reading only up to end of this chunk
							if ( (m_nTransfered + fileBytesRead + len) > m_nChunkSize )
							{
								len = m_nChunkSize - (m_nTransfered + fileBytesRead);
								//printf("Limiting len to %llu\n", len);
							}
							
							readbuffer.SetSize(len);
						
							long read = m_File.Read( (char *)readbuffer.Data(), (long)len );
						
							if ( read == -1 )
							{
								CallBack_SendError(strerror(errno));
								perror("CTransfer::DataSend() read error ! ");
								Disconnect(TRUE);
								return;
							}
							else if ( read == 0 )
							{
								perror("CTransfer::DataSend() no data read ! ");
								Disconnect(TRUE);
								return;
							}
							else
							{
								fileBytesRead += read;
								
								if ( m_nTransfered + fileBytesRead >= m_nChunkSize )
								{
									more = FALSE;
								}
								
								temp.SetSize( m_nReadFileBufferSize * 2 );
							
								inlen = read;
								outlen = temp.Size();
								
								/*
								 * Bad things will happen if the compressed data is over twice the size of
								 * the uncompressed data, fortunately that is very unlikely.
								 */
								m_nZlibStatus = m_Deflater.DeflateBlock( (char *)readbuffer.Data(), &inlen, (char *)temp.Data(), &outlen, more );
								
								if ( ((m_nZlibStatus == 1) && (more == TRUE)) ||
								     ((m_nZlibStatus == 0) && (more == FALSE)) )
								{
									if ( outlen > 0 )
									{
										pByteArray->Append( temp.Data(), outlen );
										m_nFileBufferSize += outlen;
									}
								}
								else
								{
									CString reason;
									if ( m_nZlibStatus == -1 )
									{
										reason = "zlib error";
									}
									else if ( (m_nZlibStatus == 0) && (more == TRUE) )
									{
										reason = "had more data, but zlib finished";
									}
									else if ( (m_nZlibStatus == 1) && (more == FALSE) )
									{
										/*
										 * if the compressed data was over twice the size of the uncompressed...
										 */
										reason = "no more data, but zlib not finished";
									}
									CallBack_SendError("Zlib compression failed: " + reason);
									printf("CTransfer::DataSend() deflate failed: %s\n", reason.Data());
									Disconnect(TRUE);
									return;
								}
							}
						}
					}
					
					m_nFileBufferPos = 0;
					len = m_nFileBufferSize;
					
					// must either change sizes or adjust m_nTransfered
					//SetLength( GetLength() + m_nFileBufferSize - fileBytesRead );
					//SetEndPosition( GetEndPosition() + m_nFileBufferSize - fileBytesRead );
					//m_nChunkSize = m_nChunkSize + m_nFileBufferSize - fileBytesRead;
					
					m_nTransfered = m_nTransfered + fileBytesRead - m_nFileBufferSize;
					
					//printf( "Compressed: %ld to %ld\n", fileBytesRead, m_nFileBufferSize );
				}
				else // m_nDataType == 2
				{

				m_nFileBufferSize = m_File.Read( (char *)pByteArray->Data(), (long)len );

				if ( m_nFileBufferSize == -1 )
				{
					CallBack_SendError(strerror(errno));
					perror("CTransfer::DataSend() read error ! ");
					Disconnect(TRUE);
					return;
				}
				else if ( m_nFileBufferSize == 0 )
				{
					perror("CTransfer::DataSend() no data read ! ");
					Disconnect(TRUE);
					return;
				}
				else if ( len > (ulonglong)m_nFileBufferSize )
				{
					perror("CTransfer::DataSend() wrong length calculation ! ");
					len = m_nFileBufferSize;
				}

				m_nFileBufferPos = 0;
				
				} // m_nDataType == 2
			}

			// limit to sendbuffer
			if ( len > (ulonglong)(m_nFileBufferSize-m_nFileBufferPos) )
			{
				len = m_nFileBufferSize-m_nFileBufferPos;
			}
		}
		else if ( (eMedium == eltBUFFER) || (eMedium == eltTTHL) )
		{
			m_nFileBufferPos = m_nTransfered;
		}

		// limit to transferrate
		if ( sendbyte != 0  )
		{
			if ( len > sendbyte )
			{
				len = sendbyte;
			}
		}

		// limit to free socket sendbuffer
		if ( len > freebuf )
		{
			len = freebuf;
		}

		// fix to pending send bytes
		if ( m_nPendingSendBytes != 0 )
		{
			len = m_nPendingSendBytes;
		}

//		DPRINTF("LEN: %d\n",len);

		/* For compressed buffer transfers, the entire buffer was compress in DoInitUpload*/
		// write data to socket
		l = Write( pByteArray->Data()+m_nFileBufferPos, (long)len, TRUE );

		if ( l > 0 )
		{
			AddTraffic(l);

			if ( m_nPendingSendBytes != 0 )
			{
				m_nPendingSendBytes -= l;
			}
			else
			{
				m_nPendingSendBytes = len-l;
			}

			if ( eMedium == eltFILE )
			{
				m_nFileBufferPos += l;
			}

			// add traffic control
			CSocket::m_Traffic.AddTraffic(ettDATATX,len);
		}
		else if ( l == -1 )
		{
			perror("CTransfer::DataSend() write error !\n");
			Disconnect(TRUE);
			return;
		}
		else if ( l == 0 )
		{
			if ( m_nPendingSendBytes == 0 )
			{
				m_nPendingSendBytes = len;
			}

			break;
		}

		if ( (((m_nStartPosition+m_nTransfered) == m_nFileSize) && (m_nZlibStatus == 0)) ||
		     ( ((m_MessageSupports.m_bChunk == TRUE) ||
		        (m_MessageSupports.m_bADCGet == TRUE) ||
			(m_MessageSupports.m_bXMLBZList == TRUE) ||
			(m_MessageSupports.m_bZBlock == TRUE)) &&
		       ((m_nChunkSize == m_nTransfered) && (m_nZlibStatus == 0))
		     )
		   )
		{
			DPRINTF("end found\n");
			if ( eMedium == eltFILE )
			{
				m_File.Close();
			}

			// free buffer
			pByteArray->SetSize(0);

			// stop transfer
			m_bIdle = TRUE;

			// write finished upload message into log file
			if ( (CConfig::Instance()->GetLogFile() == TRUE) && (CConfig::Instance()->GetLogFinishedUploads() == TRUE) )
			{
				CString logmessage = "Upload done [";
				logmessage += sNick;
				logmessage += "] ";
				if (sSrcFilename == "")
				{
					logmessage += "filelist";
				}
				else
				{
					logmessage += sSrcFilename;
				}
				logmessage += " (";
				logmessage += CString().setNum(m_nTransfered);
				logmessage += '/';
				logmessage += CString().setNum(m_nFileSize);
				logmessage += ')';
				CLogFile::Write(CConfig::Instance()->GetLogFileName(),eltINFO,logmessage);
			}

			// stop loop
			lo = 5;
		}

		Object = new CMessageTransfer();

		Object->m_eDirection  = edUPLOAD;
		Object->m_nTransfered = m_nStartPosition+m_nTransfered;
		Object->m_nLength     = m_nFileSize;

		CallBack_SendObject(Object);

		} // for
	}
}

/** */
void CTransfer::DataTimeout()
{
	if ( (eMode == estTRANSFERDOWNLOAD) && (m_bIdle == FALSE) )
	{
		DPRINTF("CTransfer: download timeout ...\n");

		// send string ...
		SendString("|");
	}
	else if ( eMode == estTRANSFERHANDSHAKE )
	{
		DPRINTF("CTransfer: handshake timeout ...\n");

		// handshake timeout
		Disconnect(TRUE);
	}
	else if ( m_bIdle == TRUE )
	{
		// send string ...
		SendString("|");
	}
}

/** */
void CTransfer::Notify()
{
}

/** */
void CTransfer::DataAvailable( const char * buffer, int len )
{
	int i,r,p;
	CMessageTransfer *Object;

	i = 0;
	p = 0;

	while ( i < len )
	{
		if ( (eMode == estTRANSFERDOWNLOAD) && (m_bIdle == FALSE) )
		{
			while ( (i < len) && (!m_bIdle) )
			{
				if ( eMedium == eltFILE )
				{
					r = HandleFileTransfer(buffer+i,len-i);
				}
				else if ( (eMedium == eltBUFFER) || (eMedium == eltTTHL) )
				{
					r = HandleBufferTransfer(buffer+i,len-i);
				}
				else
				{
					r = -1;
				}

				if ( r >= 0 )
				{
					i += r;
				}
				else
				{
					break;
				}

				// add traffic control
				CSocket::m_Traffic.AddTraffic(ettDATARX,r);

				// set transfer to idle on finished chunk if remote
				// support chunks or if fileend reached
				if ( ((m_nChunkSize == m_nTransfered) && (m_nZlibStatus == 0)) &&
				    ((m_MessageSupports.m_bChunk == TRUE) ||
				     (m_MessageSupports.m_bZBlock == TRUE) ||
				     (m_MessageSupports.m_bXMLBZList == TRUE) ||
				     (m_MessageSupports.m_bADCGet == TRUE) ||
				     ((m_nStartPosition+m_nChunkSize) == m_nFileSize) ) )
				{
					m_bIdle = TRUE;
					
					// send transfer message does NOT go here, moving it here
					// breaks multi download from original nmdc client microdc 0.11.0
					// which is handled by a single $Send| getting multiple chunks
				}
				
				// however, we can put the test m_nChunkSize == m_nTransfered, to
				// avoid sending transfer messages which will just be ignored
				if ( m_nChunkSize == m_nTransfered )
				{
					// send transfer message to handle new chunk-end
					Object = new CMessageTransfer();
					Object->m_eDirection  = edDOWNLOAD;
					Object->m_nTransfered = m_nTransfered;
					Object->m_nLength     = m_nFileSize;
					CallBack_SendObject(Object);
					
					// check end - only finish if zlib has finished (m_nZlibStatus is always 0 when not using compressed transfers)
					if ( m_nZlibStatus == 0 )
					{
						// if endposition not filesize and remote don't support chunks we must disconnect
						if ( ((m_nStartPosition+m_nChunkSize) != m_nFileSize) &&
					     	(m_MessageSupports.m_bChunk == FALSE) &&
					     	(m_MessageSupports.m_bZBlock == FALSE) &&
					     	(m_MessageSupports.m_bXMLBZList == FALSE) &&
					     	(m_MessageSupports.m_bADCGet == FALSE) )
						{
							Disconnect();
						}

						m_bIdle = TRUE;

						m_File.Close();
					}
				}
			}
		}
		else
		{
			// Work around a bug caused by the size of compressed data being larger
			// than the uncompressed data. This causes valknut to set the "idle" flag
			// too early, when we still are getting compressed data. So we'll check that
			// the data starts with "$", if it does then it's probably a command, otherwise
			// it's probably data. No this is not a proper fix, just makes the bug
			// much less likely to occur (since the compressed data might start with "$").
			// UPDATE: now that we check m_nZlibStatus before setting m_bIdle to TRUE
			// this horrible hack is no longer needed
			// if (buffer[i] == '$')
			//{
				r = HandleControlTransfer(buffer+i,len-i);
				
				// add traffic control
				CSocket::m_Traffic.AddTraffic(ettCONTROLRX,r);
			/* }
			else
			{
				DPRINTF("Bug avoiding! Control transfer not starting with \"$\" is really data!\n");
				if ( eMedium == eltFILE )
				{
					r = HandleFileTransfer(buffer+i,len-i);
				}
				else if ( (eMedium == eltBUFFER) || (eMedium == eltTTHL) )
				{
					r = HandleBufferTransfer(buffer+i,len-i);
				}
				else
				{
					r = -1;
				}
			} */

			if (r >= 0)
			{
				// add traffic control
				CSocket::m_Traffic.AddTraffic(ettDATARX,r);
				
				i += r;
			}
			else
			{
				break;
			}
		}

		if ( i == p )
		{
			DPRINTF("WARNING: unknown data ! [%d %d]\n",i,len);
			break;
		}
		else
		{
			p = i;
		}
	}
}

/** */
int CTransfer::HandleFileTransfer( const char * buffer, int len )
{
	int i,i1;
	ulonglong l;
	CString serr="";
	bool berr=FALSE;
	CByteArray zbuffer;
	int outlen;
	int inlen;
	
	// decompress buffer ...
	if ( m_nDataType == 1 )
	{
		zbuffer.SetSize(1024*100);
		outlen = 1024*100;
		inlen  = len;
		m_nZlibStatus = m_ZLib.InflateZBlock( (char*)buffer, &inlen, (char*)zbuffer.Data(), &outlen );
		if ( m_nZlibStatus == -1 )
		{
			DPRINTF("HandleFileTransfer: inflate failed!\n");
			serr = "Zlib decompression failed";
			
			// close file
			m_File.Close();

			CallBack_SendError(serr);

			// disconnect on error
			Disconnect();
			
			return -1;
		}
		
		len = outlen;
		buffer = (const char*)zbuffer.Data();
	}

	// check endposition
	if ( (len+m_nTransfered) > m_nChunkSize )
	{
		l = m_nChunkSize-m_nTransfered;
	}
	else
	{
		l = len;
	}

	i1 = 0;

	// write data
	if ( (m_File.IsOpen()) && (l > 0) )
	{
		while( i1 != l )
		{
			i = m_File.Write( buffer, l );

			if ( i != -1 )
			{
				i1 += i;
			}
			else
			{
				i1 -= m_File.GetBufferPos();

				DPRINTF("CTransfer::HandleFileTransfer: write failed with %d pending bytes\n",m_File.GetBufferPos());
				berr = TRUE;
				serr = strerror(errno);
				break;
			}
		}
	}
	else
	{
		// TODO: error message
	}

	AddTraffic(i1);

	// handle error
	if ( berr == TRUE )
	{
		i1 = -1;

		// close file
		m_File.Close();

		CallBack_SendError(serr);

		// disconnect on error
		Disconnect();
	}
	else
	{
		if ( m_nDataType == 1 )
		{
			i1 = inlen;
		}
	}

	return i1;
}

/** */
int CTransfer::HandleBufferTransfer( const char * buffer, int len )
{
	int i;
	CString serr="";
	bool berr=FALSE;
	CByteArray zbuffer;
	int outlen;
	int inlen;
	
	// decompress buffer ...
	if ( m_nDataType == 1 )
	{
		zbuffer.SetSize(1024*100);
		outlen = 1024*100;
		inlen  = len;
		m_nZlibStatus = m_ZLib.InflateZBlock( (char*)buffer, &inlen, (char*)zbuffer.Data(), &outlen );
		if ( m_nZlibStatus == -1 )
		{
			DPRINTF("HandleBufferTransfer: inflate failed!\n");
			serr = "Zlib decompression failed";
			
			CallBack_SendError(serr);

			// disconnect on error
			Disconnect();
			
			return -1;
		}
		
		len = outlen;
		buffer = (const char*)zbuffer.Data();
	}

	// check endposition
	if ( (len+m_nTransfered) > m_nChunkSize )
	{
		i = m_nChunkSize-m_nTransfered;
	}
	else
	{
		i = len;
	}

	pByteArray->Append((const unsigned char*)buffer,(long)i);

	AddTraffic(i);

	// end found
	/* Do not decompress filelists, let valknut handle that */
	/* if ( m_nTransfered == m_nChunkSize )
	{
		if ( sDstFilename == DC_USER_FILELIST_HE3 )
		{
			// decompress filelist
			if ( (m_MessageSupports.m_bBZList == FALSE) &&
			     (m_MessageSupports.m_bXMLBZList == FALSE) )
			{
				CHE3 *he3=0;
				CString * s=0;

				he3 = new CHE3();
				s = he3->decode_he3_data(pByteArray);
				delete he3;
				pByteArray->SetSize(0);

				if ( s != 0 )
				{
					pByteArray->Append( (unsigned char*)s->Data(), s->Length() );
					delete s;
				}
				else
				{
					berr = TRUE;
					serr = "he3 decompress failed";
				}
			}
			else
			{
				CBZ * bz = 0;
				CByteArray out;

				bz = new CBZ();

				if ( bz->Decompress( pByteArray, &out ) == TRUE )
				{
					pByteArray->SetSize(0);
					pByteArray->Append( out.Data(), out.Size() );
				}
				else
				{
					berr = TRUE;
					serr = "bz2 decompress failed";
					pByteArray->SetSize(0);
				}

				delete bz;
			}
		}
	} */

	// handle error
	if ( berr == TRUE )
	{
		i = -1;

		CallBack_SendError(serr);

		// disconnect on error
		Disconnect();
	}
	else
	{
		if ( m_nDataType == 1 )
		{
			i = inlen;
		}
	}

	return i;
}

/** */
int CTransfer::HandleControlTransfer( const char * buffer, int len )
{
	CString s;
	int i,p;

	p = 0;

	// search first '|'
	for(i=0;i<len;i++)
	{
		if ( buffer[i] == '|' )
		{
			s = sBuffer;
			s += CString().Set(buffer,i+1);
			break;
		}
	}
	
	if ( i == len )
		i = -1;
	i++;

	if ( i > 0 )
	{
		p = HandleMessage( (char*)s.Data(), i );

		// some clients send data after filelength (e.g. nmdc 2.02)
		if ( (p < s.Length()) && (!m_bIdle) && (eMode == estTRANSFERDOWNLOAD) )
		{
			DPRINTF("WARNING: transfer->client send unwanted data ! [%d %d %d]\n",i,p,len);
		}
	}

	if ( (i == s.Length()) ||
	     ((!m_bIdle) && (eMode == estTRANSFERDOWNLOAD)) )
	{
		sBuffer = "";
	}
	else
	{
		sBuffer = s.Mid( i, s.Length()-i );
		p = len;
	}

	return p;
}

/** Code copied out of HandleMessage() so it can be called directly from another place */
void CTransfer::DoInitUpload()
{
	m_nTransfered = 0;
	// init filehandle
	if ( eMedium == eltFILE )
	{
		m_nFileBufferPos    = m_nReadFileBufferSize;
		m_nFileBufferSize   = m_nReadFileBufferSize;
		m_nPendingSendBytes = 0;
		pByteArray->SetSize(m_nReadFileBufferSize);
		m_File.Close();
		
		if ( m_File.Open( sSrcFilename, IO_RAW | IO_READONLY ) == FALSE )
		{
			CallBack_SendError( CString("(File open) ") + strerror(errno) );
			perror("File open");
		}
		else
		{
			// seek to the startposition
			if ( m_File.Seek(m_nStartPosition,SEEK_SET) == FALSE )
			{
				CallBack_SendError( CString("(File seek) ") + strerror(errno) );
				perror("File seek");
				m_File.Close();
			}
		}
		// check for error
		if ( !m_File.IsOpen() )
		{
			Disconnect(TRUE);
		}
		else
		{
			// start transfer
			m_bIdle = FALSE;
		}
	}
	else
	{
		// compress buffer if necessary
		if ( m_nDataType == 2 )
		{
			m_nFileBufferPos = 0;
			CByteArray * compressed = new CByteArray();
			unsigned long outlen = m_Deflater.OneShotDeflate( (const char*)pByteArray->Data(), pByteArray->Size(), compressed );
			if ( outlen > 0 )
			{
				//printf("Compressed buffer from %ld to %ld\n", pByteArray->Size(), outlen );
				pByteArray->SetSize(0);
				pByteArray->Append( compressed->Data(), outlen );
				
				// hack
				SetEndPosition(outlen);
				SetLength(outlen);
				m_nChunkSize = outlen;
			}
			else
			{
				printf("CTransfer::DoInitUpload: Compress buffer failed!\n");
			}
			delete compressed;
		}
		
		// start transfer
		m_bIdle = FALSE;
	}
	DPRINTF("start upload ...'%s' %lld/%lld\n",sSrcFilename.Data(),m_nStartPosition,m_nChunkSize);
}

/** */
int CTransfer::prepareTTHLsend( CString tth )
{
	//DPRINTF("Get TTH leaves for %s\n", tth.Data() );
	CByteArray * leaves = CFileManager::Instance()->GetHashLeaves( tth );
	
	if (leaves != 0)
	{
		// this is what SetBuffer() does plus SetSize(0) first
		TransferMutex.Lock();
		pByteArray->SetSize(0);
		pByteArray->Append( leaves->Data(), leaves->Size() );
		
		int size = leaves->Size();
		
		m_nFileSize      = size;
		m_nStartPosition = 0;
		m_nEndPosition   = size;
		
		TransferMutex.UnLock();
		
		m_nChunkSize = 0;
		
		delete leaves;
		
		return size;
	}
	else
	{
		DPRINTF("Hash leaves not found for %s\n", tth.Data());
		SendError("File Not Available");
		Disconnect();
		return 0;
	}
}
